#
#  tkorca_controller.rb : 
#
#                             by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#
require 'thread'
require 'timeout'
require 'server/application_controller.rb'
require 'server/controller_passwd.rb'
require 'server/controller_command.rb'

TkORCA_Controller_Auth.read(SERVER_PARAM['control_passwd'])

module TkORCA_Controller
  @gate = nil
  @cmd_io_sock = nil
end

class << TkORCA_Controller
  class RECONNECT_COMMANDER   < IOError; end
  class RECONNECT_APPLICATION < IOError; end

  ####################################
  private
  ####################################
  IAC      = 255.chr # "\377"
  DONT     = 254.chr # "\376"
  DO       = 253.chr # "\375"
  WONT     = 252.chr # "\374"
  WILL     = 251.chr # "\373"
  OPT_ECHO =   1.chr # "\001"
  CR = "\015"
  LF = "\012"
  CRLF = CR + LF

  AUTH_TIMEOUT = 60
  AUTH_TRIAL = 3

  def auth_check(sock)
    user = nil
    auth = false
    begin
      sock.print("\n*** Control Port of Ruby/TkORCA Server ***\n")
      sock.flush

      timeout(AUTH_TIMEOUT){
        AUTH_TRIAL.times{|i|
          begin
            loop {sock.read_nonblock(1024)}  # clear read buffer
          rescue Exception
          end

          sock.write("Username: " + IAC + WONT + OPT_ECHO); sock.flush

          buffer = ''
          loop {
            begin
              lines = sock.read_nonblock(1024).split(/^/)
              buffer << lines.shift
              break if ! lines.empty? || buffer.chomp! || buffer.size > 1024
            rescue EOFError
              break
            rescue Exception
              sleep 0.1  # thread pass before next read
            end
          }

          buffer.gsub!(/\A.*#{IAC}[#{DO}#{DONT}]#{OPT_ECHO}/no, '')
          user = buffer.chomp.strip

          begin
            loop {sock.read_nonblock(1024)}  # clear read buffer
          rescue Exception
          end

          sock.write("Password: " + IAC + WILL + OPT_ECHO); sock.flush

          buffer = ''
          loop {
            begin
              lines = sock.read_nonblock(1024).split(/^/)
              buffer << lines.shift
              break if ! lines.empty? || buffer.chomp! || buffer.size > 1024
            rescue EOFError
              break
            rescue Exception
              sleep 0.1  # thread pass before next read
            end
          }

          buffer.gsub!(/\A.*#{IAC}[#{DO}#{DONT}]#{OPT_ECHO}/no, '')
          #passwd = $1.chomp.strip
          passwd = buffer.chomp.strip

          break if (auth = TkORCA_Controller_Auth.auth(user, passwd))
          sleep(1 + 2*i)
          sock.write(CRLF + "Invalid Username/Password ..." + CRLF + CRLF)
          sock.flush
        }
      }
    rescue Exception
      nil
    ensure
      sock.write(IAC + WONT + OPT_ECHO + CRLF)
      sock.flush
      timeout(3) { sock.readpartial(1024) }
      # str = sock.gets
      # str.gsub!(/#{IAC}[#{DO}#{DONT}]#{OPT_ECHO}/no, '')
      # str =~ /(?:#{IAC}[#{DO}#{DONT}]#{OPT_ECHO})?(.*)\z/no
      # sock.read(3) # IAC + DONT + OPT_ECHO
    end

    if auth
      sock.write("Connect ..." + CRLF + CRLF)
      sock.flush
      @controller_username = user
      LOGGER.log("User `#{@controller_username}': connect to controller")
    else
      @controller_username = nil
      sock.write("You may not have the authority. Sorry ..." + CRLF)
      sock.flush
    end
    auth
  end

  ####################################

  def reconnect_commander(sock)
    @cmd_io_sock = sock

    if @cmd_rd_thread
      begin
        @cmd_rd_thread.raise(RECONNECT_COMMANDER)
      rescue Exception
      end
    end

    if @msg_wr_thread
      begin
        @msg_wr_thread.raise(RECONNECT_COMMANDER)
      rescue Exception
      end
    end
  end

  def disconnect_cmd_io_sock
    if @cmd_io_sock
      Thread.new{put_msg("close connection\n")}
      sleep 1 # give a chance to send the message
      @cmd_io_sock.shutdown(Socket::SHUT_RDWR) rescue nil
      @cmd_io_sock.close rescue nil
    end
    reconnect_commander(nil)
    @accepter.raise(RECONNECT_COMMANDER)
    nil
  end

  def get_cmd
    if (io = @cmd_io_sock)
      @cmd_rd_thread = Thread.new{
        begin
          io.readline  # trust controller user ( == system master, probably )
        rescue RECONNECT_COMMANDER
          nil
        rescue Exception
          # cannot continue I/O operation
          disconnect_cmd_io_sock()
          nil
        end
      }
    else
      @cmd_rd_thread = Thread.new{
        begin
          sleep
        rescue RECONNECT_COMMANDER
          nil
        end
      }
    end
    @cmd_rd_thread.value
  end

  def put_msg(str)
    if (io = @cmd_io_sock)
      @msg_wr_thread = Thread.new{
        begin
          io.write(str)
          io.flush
        rescue Exception
          # cannot continue I/O operation
          disconnect_cmd_io_sock()
          nil
        end
      }
      @msg_wr_thread.value
    else
      @msg_wr_thread = nil
      nil
    end
  end

  def close_connection_buffer
    @from_app = @to_app = nil

    (@cmd_buf_rd.close rescue nil) if @cmd_buf_rd
    (@cmd_buf_wr.close rescue nil) if @cmd_buf_wr
    @cmd_buf_rd = @cmd_buf_wr = nil

    (@msg_buf_rd.close rescue nil) if @msg_buf_rd
    (@msg_buf_wr.close rescue nil) if @msg_buf_wr
    @msg_buf_rd = @msg_buf_wr = nil

    nil
  end

  def reconnect_application
    if @from_app_thread
      begin
        @from_app_thread.raise(RECONNECT_APPLICATION)
      rescue Exception
      end
    end

    if @to_app_thread
      begin
        @to_app_thread.raise(RECONNECT_APPLICATION)
      rescue Exception
      end
    end
  end

  def connect_application(id)
    if (app = TkORCA_Service::Connection_ID_table[id])
      close_connection_buffer()

      @cmd_buf_rd, @cmd_buf_wr = IO.pipe
      @msg_buf_rd, @msg_buf_wr = IO.pipe
      @from_app = @msg_buf_rd
      @to_app   = @cmd_buf_wr

      @application = app

      begin
        app.connect(@cmd_buf_rd, @msg_buf_wr)
        reconnect_application()
      rescue => e
        close_connection_buffer()
        put_msg("#{e.class}: #{e.message}\n")
      end
    else
      put_msg("Error: no running application which has ID == #{id}\n")
    end
  end

  def disconnect_application
    close_connection_buffer()

    if (app = @application)
      @application = nil
      begin
        put_msg("disconnect application\n")
      rescue Exception
      end
      app.disconnect
    end

    reconnect_application()
    nil
  end

  ####################################

  def show_prompt
    put_msg("Pod[#{SERVER_PARAM['port']}]> ")
  end

  def show_control_command_help(cmd_str, cmd, help_target=nil)
    if help_target == "?"
      put_msg("? [command] : show command list, or help of the command\n")
    elsif help_target
      operation, help_msg = 
        TkORCA_Controller::CONTROL_COMMAND_TABLE[help_target]
      if operation
        if help_msg
          put_msg("#{help_msg}\n")
        else
          put_msg("`#{help_target}' has no help message >\n")
        end
      else
        put_msg("unknown command `#{help_target}'\n")
      end
    else
      put_msg("?\n" + 
                TkORCA_Controller::CONTROL_COMMAND_TABLE.keys.sort.join("\n") +
                "\n")
    end
  end

  def control_pod_leader(cmd)
    return nil unless cmd

    cmd_str = cmd.chomp.strip
    return true if cmd_str.empty?

    cmd_line = cmd_str.split(' ')
    if (cmd_line[0] == '?')
      show_control_command_help(cmd_str, *cmd_line)
      true
    elsif (cmd_info = TkORCA_Controller::CONTROL_COMMAND_TABLE[cmd_line[0]])
      operation, help_msg = *cmd_info
      begin
        ((ret = operation.call(cmd_str, *cmd_line)) == false)? nil: ret
      rescue Exception => e
        put_msg("#{e.class}: #{e.message}\n")
        nil
      end
    else # not control command
      unless @application
        put_msg("unknown command `#{cmd_line[0]}'\n")
        true
      else
        false
      end
    end
  end

  def control_operation
    @operation_thread = Thread.new{
      if (cmd = get_cmd()).nil?
        false
      elsif (ret = control_pod_leader(cmd)) != false
        ret
      else
        if (io = @to_app)
          @to_app_thread = Thread.new{
            begin
              io.write(cmd)
              io.flush
              false
            rescue RECONNECT_APPLICATION
              true
            rescue Exception
              # cannot continue I/O operation
              #put_msg("error on the connection to application\n")
              disconnect_application()
              true
            end
          }
          @to_app_thread.value
        else
          # no connect to application
          @to_app_thread = nil
          true
        end
      end
    }
    prompt_mode = @operation_thread.value
    show_prompt() if prompt_mode
    nil
  end

  def forward_message
    if (io = @from_app)
      @from_app_thread = Thread.new{
        begin
          put_msg(io.readpartial(1024))
        rescue RECONNECT_APPLICATION
          nil
        rescue Exception
          # cannot continue I/O operation
          #put_msg("error on the connection to application\n")
          disconnect_application()
          show_prompt()
          nil
        end
      }
    else
      @from_app_thread = Thread.new{
        begin
          sleep
        rescue RECONNECT_APPLICATION
          nil
        end
      }
    end
    @from_app_thread.value
  end

  def create_gate
    # host = 'localhost'
    host = '127.0.0.1'

    begin
      @gate = TCPServer.open(host, SERVER_PARAM['control_port'])

    rescue Exception => e
      LOGGER.log("ERROR: #{e.class}: #{e.message}")
      LOGGER.log("fail to create Ruby/TkORCA control port :", 
                 "host = `#{host}', port = #{SERVER_PARAM['control_port']}")
      raise e

    else
      LOGGER.log("create Ruby/TkORCA control port :", 
                 "host = `#{host}', port = #{SERVER_PARAM['control_port']}")
    end
  end

  def create_connection
    begin
      @controller_username = nil
      if auth_check(sock = @gate.accept)
        reconnect_commander(sock)
        show_prompt()
        sleep
      else
        sock.shutdown(Socket::SHUT_RDWR) rescue nil
        sock.close rescue nil
      end
    rescue RECONNECT_COMMANDER
      nil
    rescue Exception => e
      LOGGER.log("ERROR: #{e.class}: #{e.message}")
      LOGGER.log("fail to start Ruby/TkORCA application controller", 
                 "for #{@cmd_io_sock.peeraddr.inspect}")
    ensure
      sock.shutdown(Socket::SHUT_RDWR) rescue nil
      sock.close rescue nil

      if @controller_username
        LOGGER.log "User `#{@controller_username}': disconnect from controller"
      end
      @controller_username = nil
    end
  end


  ####################################
  public
  ####################################

  def start
    return if @gate && ! @gate.closed?  # already exist
    @gate = nil

    if @cmd_io_sock
      disconnect_cmd_io_sock()
    else
      @cmd_io_sock = nil
    end

    @cmd_rd_thread = nil
    @msg_wr_thread = nil

    @application = nil

    @from_app = nil
    @to_app   = nil

    @from_app_thread = nil
    @to_app_thread   = nil

    @cmd_buf_rd = @cmd_buf_wr = nil
    @msg_buf_rd = @msg_buf_wr = nil

    @sender   = Thread.new{ loop { control_operation() } }
    @receiver = Thread.new{ loop { forward_message()   } }

    create_gate()
    @accepter = Thread.new{ loop { create_connection() } }
  end

  def restart
    kill()
    start()
  end

  def kill
    disconnect_application()
    disconnect_cmd_io_sock()
    close_connection_buffer()
    (@gate.shutdown(Socket::SHUT_RDWR) rescue nil) if @gate && !@gate.closed?
    @accepter.kill rescue nil
    @gate.close rescue nil
    nil
  end
  alias stop kill
end
