#
#  tkorca_service.rb : 
#
#                             by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#
require 'thread'
require 'server/new_display_num.rb'
require 'server/create_temp_home.rb'
require 'server/create_xauth.rb'
require 'server/xvnc_process.rb'
require 'server/application_process.rb'
require 'server/tkorca_controller.rb'

class TkORCA_Service
  include Free_Display_Num
  include Temporal_Home_Handler
  include Xauth_File_Handler
  include Xvnc_Process_Handler
  include Application_Process_Handler

  @@mutex = Mutex.new
  @@next_connection_ID = 0
  Connection_ID_table  = {}
  Xvnc_process_table   = {}
  TkORCA_process_table = {}

   WITH_RUBY_VM = Object.const_defined?(:VM) && ::VM.class == Class

  if WITH_RUBY_VM
    @@xauth_file_mutex  = Mutex.new
    @@sel_display_mutex = Mutex.new
  end

  ########################

  def self.kill_all_connections
    table = nil
    @@mutex.synchronize{ table = Xvnc_process_table.dup }
    table.each{|pid, service| service.kill }
  end

  ########################

  def _get_lock_file_name
    xauth_dir = get_xauth_dir()
    File.join(xauth_dir, 'lock_for_checking_display')
  end

  def _get_connection_id
    new_id = -1
    @@mutex.synchronize{
      @@next_connection_ID = 0 if @@next_connection_ID >= 10000
      while Connection_ID_table.has_key?(@@next_connection_ID)
        @@next_connection_ID += 1
      end
      new_id = @@next_connection_ID
      @@next_connection_ID += 1
    }
    new_id
  end
  private :_get_connection_id

  def _remove_xauth_file
    if (file = @xauth_file)
      begin
        if WITH_RUBY_VM
          @@xauth_file_mutex.lock
        else
          old_crit = Thread.critical
          Thread.critical = true
        end
        File.delete(file) if FileTest.exist?(file)
      rescue
        LOGGER.log("ID[#{@connection_id}] fail to remove xauth-file", 
                   "`#{File.basename(file)}'")
      else
        LOGGER.log("ID[#{@connection_id}] remove xauth-file", 
                   "`#{File.basename(file)}'")
        @xauth_file = nil
      ensure
        if WITH_RUBY_VM
          @@xauth_file_mutex.unlock
        else
          Thread.critical = old_crit
        end
      end
    end
  end
  private :_remove_xauth_file

  def _remove_temp_home_dir
    begin
      if @temp_home && FileTest.exist?(@temp_home)
        remove_temporal_home(@temp_home)
      end
    rescue Exception => e
      LOGGER.log("ID[#{@connection_id}] fail to remove `#{@temp_home}'." + 
                   " please remove it by hand. ;" + 
                   " ERROR: #{e.class}: #{e.message}")
    ensure
      @temp_home = nil
    end
  end
  private :_remove_temp_home_dir

  def _kill_tkorca_process
    begin
      if (pid = @tkorca_pid)
        @tkorca_pid = nil
        LOGGER.log("ID[#{@connection_id}] kill TkORCA process #{pid}")
        Process.kill(:TERM, pid)   # SIGTERM is first
        Thread.new{                # next SIGKILL 
          sleep 3                  # delay for termination of application
          Process.kill(:KILL, pid)
        }
      end
    rescue Exception
      # ignore
    end

    @@mutex.synchronize{ TkORCA_process_table.delete(pid) if pid }
    pid
  end
  private :_kill_tkorca_process

  def _kill_xvnc_process
    begin
      if (pid = @xvnc_pid)
        @xvnc_pid = nil
        LOGGER.log("ID[#{@connection_id}] kill Xvnc process #{pid}")
        Process.kill(SERVER_PARAM['hangup_signal'], pid) # SIGHUP is first
        Process.kill(:TERM, pid)   # next SIGTERM
        Thread.new{                # next SIGKILL 
          sleep 3                  # delay for termination of application
          Process.kill(:KILL, pid)
        }
      end
    rescue Exception
      # ignore
    end

    _remove_xauth_file()
    @@mutex.synchronize{ Xvnc_process_table.delete(pid) if pid }
    _remove_temp_home_dir()

    pid
  end
  private :_kill_xvnc_process

  def _create_tkorca_shutdown_thread
    Thread.new{
      pid = @tkorca_pid
      begin
        if pid
          ret = Process.waitpid(pid)
          @@mutex.synchronize{ TkORCA_process_table.delete(pid) if pid }
          LOGGER.log("ID[#{@connection_id}] TkORCA application (#{pid}) exit", 
                     ": status = #{ret[1].inspect}")
          @tkorca_pid = nil
        else
          LOGGER.log("ID[#{@connection_id}] No TkORCA application process")
        end
      ensure
        # Important to close connection
        _kill_xvnc_process()
        _remove_temp_home_dir()
      end
    }
  end
  private :_create_tkorca_shutdown_thread

  def _create_xvnc_shutdown_thread
    Thread.new{
      pid = @xvnc_pid
      begin
        if pid
          ret = Process.waitpid(pid)
          @@mutex.synchronize{ Xvnc_process_table.delete(pid) if pid }
          LOGGER.log("ID[#{@connection_id}] Xvnc (#{pid}) exit", 
                     ": status = #{ret[1].inspect}")
          LOGGER.log("ID[#{@connection_id}] close connection #{@peer_info}")
          @xvnc_pid = nil
        end

      ensure
        _remove_xauth_file()
        @@mutex.synchronize{
          app = Connection_ID_table.delete(@connection_id)
          app.disconnect if app
        }
        _remove_temp_home_dir()
      end
    }
  end
  private :_create_xvnc_shutdown_thread

  ########################

  def initialize(sock)
    @controller = nil

    @connection_id = _get_connection_id
    @@mutex.synchronize{ Connection_ID_table[@connection_id] = self }

    mod_opts = {}
    if SERVER_PARAM['rfbauth_passwd']
      mod_opts['rfbauth'] = SERVER_PARAM['rfbauth_passwd']
    end
    if SERVER_PARAM['desktop_name']
      mod_opts['desktop'] = SERVER_PARAM['desktop_name']
    end
    if SERVER_PARAM['wall_size']
      mod_opts['geometry'] = SERVER_PARAM['wall_size']
    end
    if SERVER_PARAM['color_depth']
      mod_opts['depth'] = SERVER_PARAM['color_depth']
    end
    if SERVER_PARAM['deferupdate']
      mod_opts['deferupdate'] = SERVER_PARAM['deferupdate']
    end
    if SERVER_PARAM['font_path']
      mod_opts['fp'] = SERVER_PARAM['font_path']
    end

    @startup_time = Time.now

    LOGGER.log("create a TkORCA service for Connection ID = #{@connection_id}")

    @peer_info = sock.peeraddr.inspect
    LOGGER.log("ID[#{@connection_id}] connection from #{@peer_info}")

    lockfile = _get_lock_file_name()
    lock_fd = open(lockfile, 'w')
    begin
      # critical part for selecting display-number and starting service
      if WITH_RUBY_VM
        @@sel_display_mutex.lock
      else
        old_crit = Thread.critical
        Thread.critical = true
      end
      lock_fd.flock(File::LOCK_EX)
      begin
        lock_fd.truncate(0)
        lock_fd.write("#{Process.pid.to_s}\n")
        lock_fd.flush
      rescue
      end

      if SERVER_PARAM['use_tmp_home']
        @temp_home = create_temporal_home()
      else
        @temp_home = nil
      end

      @display_num = get_free_display_num()
      LOGGER.log("ID[#{@connection_id}] use display :#{@display_num}")

      @xauth_file = create_xauth_file(@display_num)

      # create a Xvnc screen
      @xvnc_pid = fork_xvnc(@display_num, @xauth_file, sock, mod_opts)
      @@mutex.synchronize{ Xvnc_process_table[@xvnc_pid] = self }
      LOGGER.log("ID[#{@connection_id}] start Xvnc: process-ID = #{@xvnc_pid}")
      _create_xvnc_shutdown_thread

      10.times{ break unless is_free_display_num?(@display_num); sleep 0.5 }

      sock.close

      if (is_free_display_num?(@display_num))
        LOGGER.log("ID[#{@connection_id}] " + 
                     "fail to open display :#{@display_num}")
        _kill_xvnc_process()
        return
      end

      begin
        logger_r,   logger_w   = IO.pipe  # LOGGER
        to_irb_r,   to_irb_w   = IO.pipe  # TO IRB
        from_irb_r, from_irb_w = IO.pipe  # FROM_IRB
      rescue => e
        LOGGER.log("ID[#{@connection_id}] fail to create pipes for connection")
        _kill_xvnc_process()
        return
      end

      # create a tkorca family
      @tkorca_pid = fork_tkorca(@display_num, @xauth_file, @temp_home, 
                                [ [logger_r,   logger_w], 
                                  [to_irb_r,   to_irb_w], 
                                  [from_irb_r, from_irb_w] ])
      @@mutex.synchronize{ TkORCA_process_table[@tkorca_pid] = self }
      LOGGER.log("ID[#{@connection_id}] TkORCA process = #{@tkorca_pid}")
      _create_tkorca_shutdown_thread

      logger_w.close   rescue nil
      to_irb_r.close   rescue nil
      from_irb_w.close rescue nil

      to_irb_w.sync = true

      @controller = Application_Controller.new(@tkorca_pid, 
                                               to_irb_w, from_irb_r, logger_r)
    ensure
      lock_fd.flock(File::LOCK_UN)
      if WITH_RUBY_VM
        @@sel_display_mutex.unlock
      else
        Thread.critical = old_crit
      end
      lock_fd.close rescue nil
    end

    LOGGER.log("ID[#{@connection_id}] startup complete")
  end

  def connect(get_cmds, put_msg)
    @controller.connect(get_cmds, put_msg) if @controller
  end

  def disconnect
    @controller.disconnect if @controller
  end

  def kill
    _kill_tkorca_process()
    _kill_xvnc_process()
    @controller.kill
    _remove_temp_home_dir()
    self
  end

  attr_reader :connection_id, :startup_time, :display_num, :peer_info
end
