#!/usr/bin/env ruby
#
#  Usage: tkorca.rb [ ... options ... ] [--] [ ... applications ... ]
#
#                             by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#
##############################################
#  load version information
##############################################
load File.join(File.dirname(__FILE__), 'version.rb')


##############################################
#  create root module
##############################################
module TkORCA
  STATUS = 'running'
end


##############################################
#  get current display
##############################################
DISPLAY = ENV['DISPLAY']


##############################################
#  add SYSTEM_DIR to load-path
##############################################
SYSTEM_DIR = File.dirname(File.expand_path(__FILE__))
$LOAD_PATH.unshift(SYSTEM_DIR)


##############################################
#  setup base definitions
##############################################
# base_defs = File.join(SYSTEM_DIR, 'tkorca_base_defs')
# load(base_defs) if FileTest.readable?(base_defs)
load(File.join(SYSTEM_DIR, 'tkorca_base_defs'))


##############################################
#  define required defaults
##############################################
unless Object.const_defined?(:SYSCONF_DIR)
  # SYSCONF_DIR = SYSTEM_DIR
  SYSCONF_DIR = File.join(SYSTEM_DIR, 'sysconf')
end

unless Object.const_defined?(:SCRIPT_DIR)
  # SCRIPT_DIR = SYSTEM_DIR
  SCRIPT_DIR = File.join(SYSTEM_DIR, 'scripts')
end

unless Object.const_defined?(:AUTHORITY_CHECK)
  AUTHORITY_CHECK = File.join(SYSCONF_DIR, 'authority_check.rb')
end

unless Object.const_defined?(:DEFAULT_AUTH_MODE)
  DEFAULT_AUTH_MODE = {
    :authority => false, 
    :allow_trusted_slave => false, 
    :allow_safelevel => 4, 
    :user_name => ''
  }
end

unless Object.const_defined?(:APPLICATION_SELECTOR)
  APPLICATION_SELECTOR = File.join(SCRIPT_DIR, 'application_selector.rb')
end

unless Object.const_defined?(:SYSTEM_RCFILE)
  SYSTEM_RCFILE = File.join(SYSCONF_DIR, 'tkorcarc')
end

unless Object.const_defined?(:LOCAL_RCFILE)
  LOCAL_RCFILE = '~/.tkorcarc'
end

unless Object.const_defined?(:SYSTEM_WALLPAPER)
  SYSTEM_WALLPAPER = File.join(SYSCONF_DIR, 'tkorca-background.gif')
  # SYSTEM_WALLPAPER = File.join(SYSTEM_DIR, 'tkorca-background.gif')
  # SYSTEM_WALLPAPER = File.join(SYSTEM_DIR, 'tkorca-background2.gif')
  # SYSTEM_WALLPAPER = File.join(SYSTEM_DIR, 'tkorca-background3.gif')
end

unless Object.const_defined?(:WALL_CONFIGURE)
  WALL_CONFIGURE = nil
end

unless Object.const_defined?(:DEFAULT_TRUST_MODE_WIDTH)
#  DEFAULT_TRUST_MODE_WIDTH = 800
  DEFAULT_TRUST_MODE_WIDTH = 768
end
unless Object.const_defined?(:DEFAULT_TRUST_MODE_HEIGHT)
#  DEFAULT_TRUST_MODE_HEIGHT = 600
  DEFAULT_TRUST_MODE_HEIGHT = 576
end


##############################################
#  get command-line options
##############################################
# load(File.join(SYSTEM_DIR, 'tkorca', 'opt_parse.rb'))
require 'tkorca/opt_parse'
# p OPTION


##############################################
#  show version ?  -->  print and exit
##############################################
if OPTION['version']
  load(File.join(SYSTEM_DIR, 'show_version.rb'))
  load(File.join(SYSTEM_DIR, 'show_env_version.rb'))
  Kernel.exit(0)
end


########################################################
#  hangup signal
########################################################
if OPTION['hangup_signal']
  TkORCA::SIGHUP = OPTION['hangup_signal'].sub(/^SIG/, '')
elsif Object.const_defined?(:TkORCA_SIGHUP)
  TkORCA::SIGHUP = TkORCA_SIGHUP.sub(/^SIG/, '')
else
  list = Signal.list
  if list.has_key?('HUP')
    TkORCA::SIGHUP = :HUP
  elsif list.has_key?('BREAK')
    TkORCA::SIGHUP = :BREAK
  else
    raise RuntimeError, "Fail to select proper singal for TkORCA::SIGHUP." + 
      " It is required for interactive control." + 
      " Please define TkORCA_SIGHUP on `tkorca_base_defs' file" + 
      " or use `-S' (or `--hangup_signal') option."
  end
end


##############################################
#  setup ENV['HOME']
##############################################
if OPTION['home_dir']
  ENV['HOME'] = OPTION['home_dir']
  Dir.chdir()
end


##############################################
#  setup output for logging
##############################################
require 'tkorca/client_logger'

prefix = (OPTION['log_prefix'])? OPTION['log_prefix'].to_s: ''
if OPTION['log_fileno'].kind_of?(Fixnum) && OPTION['log_fileno'] > 0
  TkORCA::LOG = TkORCA::CLIENT_LOGGER.new(OPTION['log_fileno'], prefix)
else
  TkORCA::LOG = TkORCA::CLIENT_LOGGER.new($stderr.fileno, prefix)
end


##############################################
#  load system_rcfile (if exist)
# ( Maybe, depends on ENV['HOME'] )
##############################################
if SYSTEM_RCFILE
  rcfile = File.expand_path(SYSTEM_RCFILE)
  load(rcfile) if FileTest.readable?(rcfile)
end


##############################################
#  load local_rcfile (if needed)
# ( Maybe, depends on ENV['HOME'] )
##############################################
rcfile = nil
if OPTION['local_rcfile']
  rcfile = File.expand_path(OPTION['local_rcfile'])
elsif LOCAL_RCFILE
  rcfile = File.expand_path(LOCAL_RCFILE)
end
load(rcfile) if rcfile && FileTest.readable?(rcfile)


##############################################
#  set window size when trust-mode
##############################################
OPTION['width']  = DEFAULT_TRUST_MODE_WIDTH  if OPTION['width']  < 0
OPTION['height'] = DEFAULT_TRUST_MODE_HEIGHT if OPTION['height'] < 0


##############################################
#  add SCRIPT_DIR to load-path
# ( Maybe, depends on ENV['HOME'] )
##############################################
$LOAD_PATH << File.expand_path(SCRIPT_DIR)


##############################################
#  disable X-INPUT-METHOD ?
##############################################
if OPTION['disable_xim']
  TkORCA::USE_INPUT_METHOD = nil
end


##############################################
#  initialize
##############################################
require 'tkorca/initialize'

# Now, the mother's root widget exists.
def TkORCA.exit
  # exit with finalize operation
  Thread.new{
    TkORCA::STATUS.replace 'aborting'
    # Thread.new{Tk.root.destroy}
    TkORCA::EVLOOP_THREAD.kill
    Tk.exit rescue nil
  }
end


##############################################
#  setup debug mode
##############################################
def _setup_debug_mode
  TkORCA::DEBUG.debug_mode = true
  TkORCA::IO_WRAPPER.wrap_stdout
  TkORCA::IO_WRAPPER.wrap_stderr
end

_setup_debug_mode() if OPTION['debug_mode']
TkORCA::DEBUG.trust_mode = true  if OPTION['trust_mode']


##############################################
#  is BROADBAND network ?
##############################################
if OPTION['broadband']
  class TkORCA::RFB_Canvas
    @@broadband_network = true
  end
#  class TkORCA::Window_Frame
#    @@broadband_network = true
#  end
end


##############################################
#  start a mother (with initialization)
##############################################
MOTHER = mother = TkORCA::Mother.new(OPTION['width'], OPTION['height'])
wall = mother.wall

if OPTION['vroot_width'] > 0 || OPTION['vroot_height'] > 0
  w, h = wall.get_wallsize
  w = OPTION['vroot_width']  if OPTION['vroot_width']  > 0
  h = OPTION['vroot_height'] if OPTION['vroot_height'] > 0
  wall.set_wallsize(w, h)
end

mother.reset_timeout_timer(OPTION['timeout'])


##############################################
#  wall configure (if needed)
##############################################
wall.configure(WALL_CONFIGURE) if WALL_CONFIGURE.kind_of?(Hash)


##############################################
#  set wall-paper
##############################################
img = nil
bg_color = nil
if OPTION['wall_image']
  if OPTION['wall_image'] =~ /\A#[0-9A-Fa-f]{6}\Z/  ###
    # background color
    bg_color = OPTION['wall_image']
  else
    # wall-paper image
    img = TkPhotoImage.new(:file=>File.expand_path(OPTION['wall_image']))
  end
elsif SYSTEM_WALLPAPER
  if SYSTEM_WALLPAPER =~ /\A#[0-9A-Fa-f]{6}\Z/  ###
    # background color
    bg_color = SYSTEM_WALLPAPER
  else
    # wall-paper image
    img = TkPhotoImage.new(:file=>File.expand_path(SYSTEM_WALLPAPER))
  end
end
wall.wallbackground(bg_color) if bg_color
wall.put_wallpaper(img) if img


##############################################
#  load mother_rcfile (if needed)
##############################################
load(File.expand_path(OPTION['mother_rcfile'])) if OPTION['mother_rcfile']


##############################################
#  write initial message on the wall
##############################################
begin
  # load(File.join(SYSTEM_DIR, 'tkorca', 'default_wall_msg.rb'))
  load(File.join(SYSCONF_DIR, 'init_wall_msg.rb'))
rescue Exception
  # not important -> ignore error
end


##############################################
#  setup console
##############################################
unless OPTION['interactive']
  TkORCA::IO_WRAPPER.wrap_stdin
  TkORCA::IO_WRAPPER.wrap_stdout
  TkORCA::IO_WRAPPER.wrap_stderr
end

$stdin  = TkORCA::Console.new($stdin)
$stdout = TkORCA::Console.new($stdout)
$stderr = TkORCA::Console.new($stderr)

if OPTION['sys_console']
  sys_console = TkORCA::Console.create()
  TkORCA::DEBUG.debug_out = TkORCA::Console.console_obj
  Tk.update
  sys_console.geometry('-0+0')
  sys_console.iconify
end


##############################################
#  replace $0 and $PROGRAM_NAME
##############################################
load 'tkorca/replace_progname_globalvar.rb'


##############################################
#  start interactive-mode ( if required )
##############################################
if OPTION['interactive']
  # if defined `log_prefix', use it to make prompt
  TkORCA::IRB_PREFIX = OPTION['log_prefix'] if OPTION['log_prefix']
  load 'tkorca/interactive_mode.rb'
else
  TkORCA::IRB_THREAD = nil
end


##############################################
#  signal handling
##############################################
Signal.trap(TkORCA::SIGHUP){
  if TkORCA::IRB_THREAD
    $stdout.print "\n"
    TkORCA::IRB_THREAD.raise RubyLex::TerminateLineInput
  else
    mother._signal_handler('HUP')
  end
}

Signal.trap(:INT){
  if TkORCA::IRB_THREAD 
    st = TkORCA::IRB_THREAD['SIGINT_trap'].receive?
    if st
      $stdout.print "^C\n"
      TkORCA::IRB_THREAD.raise RubyLex::TerminateLineInput
    elsif st.nil?
      mother._signal_handler(:INT)
    else
      # ignore signal
    end
  else
    mother._signal_handler(:INT)
  end
}

Signal.trap(:TERM){mother._signal_handler(:TERM)}


##############################################
#  authentication
##############################################
AUTH_STATUS = DEFAULT_AUTH_MODE.dup
AUTH_STATUS[:authority] = false

if OPTION['auth_check']
  TkORCA::CURRENT_STAGE.replace 'authentication of tkorca'

  begin
    if FileTest.readable?(AUTHORITY_CHECK)
      load(AUTHORITY_CHECK)
    else

      STDERR.puts "Error: cannot read " + 
        "the AUTHORITY_CHECK script file: `#{AUTHORITY_CHECK}'"
      TkORCA.exit
    end

    unless AUTH_STATUS[:authority]
      STDERR.puts "fail to login"
      TkORCA.exit  # exit with finalize operation
    end
  rescue Exception => e
    STDERR.puts "#{e.class}: #{e.message}"
    e.backtrace.each{|bt| STDERR.puts "\t#{bt}"}

    TkORCA.exit  # exit with finalize operation
  end
else
  AUTH_STATUS[:authority] = true
end


##############################################
#  select applications
##############################################
TkORCA::CURRENT_STAGE.replace 'select applications'

require 'tkorca/parse_app_opt'

# Constants for the application selector.
# They will be removed after returning from the selector.
STARTUP_APPLICATIONS = []

begin
  if ! AUTH_STATUS[:authority]
    STARTUP_APPLICATIONS.clear

  elsif FileTest.readable?(APPLICATION_SELECTOR)
    load(APPLICATION_SELECTOR)

  else
    STDERR.puts "Warning: cannot read " + 
      "the APPLICATION_SELECTOR script file: `#{APPLICATION_SELECTOR}'"
    ARGV.each{|app|
      STARTUP_APPLICATIONS << 
        TkORCA::Parse_Application_Opt.parse_app_opt(app)
    }
  end

  if STARTUP_APPLICATIONS.empty?
    if AUTH_STATUS[:authority]
      STDERR.puts "There is no application for running..."
    end
    TkORCA.exit  # exit with finalize operation
  end
rescue Exception => e
  STDERR.puts "#{e.class}: #{e.message}"
  e.backtrace.each{|bt| STDERR.puts "\t#{bt}"}

  STARTUP_APPLICATIONS.clear
  TkORCA.exit  # exit with finalize operation
end

# after here, disable init_proc paramater for security reason
TkORCA::Parse_Application_Opt.disable_init_proc


##############################################
#  start daughters
##############################################
TkORCA::CURRENT_STAGE.replace 'run startup applications'

# status check
if TkORCA::STATUS == 'running'
  startup_applications = STARTUP_APPLICATIONS
else
  startup_applications = []
end

# remove constants (for security reason)
class Object
  remove_const(:MOTHER)
  remove_const(:AUTH_STATUS)
  remove_const(:STARTUP_APPLICATIONS)
end
#startup_applications << 
#  ['/home/nagai/RubyTkORCA/scripts/wrap-test-target.rb', false, 4, nil, nil]

# run applications on each daughter
startup_applications.each_with_index{|info, idx|
  script, trust, safe, console, init = info

  TkORCA::LOG.log("start new daughter:: startup_app_idx=#{idx}" \
                  + ", script=`#{script}'" \
                  + ", trust-mode=#{trust.inspect}" \
                  + ", safe-level=#{safe}" \
                  + ", use-console=#{console.inspect}" \
                  + ", init-script=#{init.inspect}" \
                  )

  daughter = mother.new_daughter(trust, safe, init)

  #daughter._invoke('wm', 'close_button_notify', '.', 
  #                 'Really exit this application?')

  if console
    daughter.__eval_proc__(console){|size|
      begin
        cons = TkORCA::Console.create('Console')
        cons.font.size = size if size.kind_of?(Fixnum)
        cons.lower
=begin
        if size.kind_of?(Fixnum)
          TkORCA::Console.create('Console', :font=>"courier #{size.inspect}").lower
        else
          TkORCA::Console.create('Console').lower
        end
=end
      rescue => e
        TkORCA::DEBUG.p [e, e.backtrace]
      end
    }
  end

  script.untaint
  daughter.load_script(script)
}

##############################################
#  wait for exit
##############################################
begin
  TkORCA::CURRENT_STAGE.replace 'wait for exit'
  TkORCA::EVLOOP_THREAD.join

ensure
  TkORCA::CURRENT_STAGE.replace 'finalize'

  TkORCA::IRB_THREAD.kill if TkORCA::IRB_THREAD

  if TkORCA::Input_Method::PID
    # stop the input method process
    begin
      Process.kill(:KILL, TkORCA::Input_Method::PID)
    rescue Errno::EXXX
    end
    begin
      Process.kill(:TERM, TkORCA::Input_Method::PID)
    rescue Errno::EXXX
    end
  end
end

# abort ?
Kernel.exit!(1) if TkORCA::STATUS == 'aborting'

# Process.kill(:KILL, VNCCONFIG)
TkORCA::STATUS.replace 'dead'
TkORCA::CURRENT_STAGE.replace 'exit'

##############################################
#  end of Ruby/TkORCA
##############################################
TkORCA::LOG.flush
Kernel.exit(0)
