#
#  tkorca/debug.rb : utility methods for debugging
#
#                           by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#
if $0 == __FILE__
  $LOAD_PATH << File.expand_path(File.join(File.dirname(__FILE__), '..'))
end

require 'tkorca/utils'

module TkORCA
  module DEBUG
    class << self
      allows = ['__id__', '__send__', 'object_id', 'to_s', 'freeze']
      undef_method(*(instance_methods - allows - allows.map{|s| s.to_sym}))
    end

    ###########################

    @debug_mode = [false, false]
    @debug_out = [STDERR]

    def self.debug_mode
      @debug_mode[0]
    end

    def self.debug_mode=(mode)
      TkORCA.security_check(1)  # allow only when $SAFE==0 
      @debug_mode[0] = (mode)? true: false
    end

    def self.trust_mode
      @debug_mode[1]
    end

    def self.trust_mode=(mode)
      TkORCA.security_check(1)  # allow only when $SAFE==0 
      @debug_mode[1] = (mode)? true: false
    end

    def self.debug_out=(out)
      TkORCA.security_check(1)  # allow only when $SAFE==0 

      begin
        @debug_out[0].flush if @debug_out[0]
      rescue Exception
        # ignore
      end

      if ! out || out.respond_to?(:write)
        @debug_out[0] = out
      else
        fail TypeError, 'debug_out must have write method'
      end
    end

    def self.kill
      @writer.kill if @writer
      @debug_out[0] = nil
    end

    ###########################

    @kernel_to_s = Kernel.instance_method(:to_s)
    @kernel_kind_of = Kernel.instance_method(:kind_of?)

    @hash_to_a = Hash.instance_method(:to_a)
    @array_collect = Array.instance_method(:collect)
    @string_length = String.instance_method(:length)

    queue = Queue.new

    @writer = Thread.new{
      loop{
        begin
          str = queue.deq
          if @debug_out[0] && @kernel_kind_of.bind(str).call(String)
            @debug_out[0].write(str)
          end
        rescue Exception => e
          # STDERR.puts e
        ensure
          begin
            @debug_out[0].flush if @debug_out[0]
          rescue Exception
            # ignore
          end
        end
      }
    }

    @wrapper = proc{|str|
      if @debug_mode[0]
        begin
          # @debug_out[0].write("#{MultiTkIp.__getip.inspect}: #{str}")
          queue.enq("#{MultiTkIp.__getip.inspect}: #{str}")
        rescue Exception=>e
          # STDERR.puts e
        end
      end
    }.freeze

    @inspector = proc{|obj, pp_mode|
      if obj == nil
        (pp_mode)? 'nil': ''
      elsif @kernel_kind_of.bind(obj).call(Hash)
        ary = proc{$SAFE=4; @hash_to_a.bind(obj).call}.call.taint
        ((pp_mode)? '{': '') << proc{
          $SAFE=4
          @array_collect.bind(ary).call{|k,v|
            "#{@inspector.call(k, pp_mode)}=>#{@inspector.call(v, pp_mode)}"
          }
        }.call.join((pp_mode)? ', ': '') << ((pp_mode)? '}': '')
      elsif @kernel_kind_of.bind(obj).call(Array)
        ((pp_mode)? '[': '') << proc{
          $SAFE=4
          @array_collect.bind(obj).call{|item| @inspector.call(item, pp_mode)}
        }.call.join((pp_mode)? ', ': '') << ((pp_mode)? ']': '')
      else
        begin
          begin
            unless obj.tainted?
              dup_obj = obj.dup.taint
            else
              dup_obj = obj
            end
          rescue TypeError
            dup_obj = obj
          end
          proc{
            $SAFE=4
            if pp_mode
              dup_obj.inspect
            else
              dup_obj.to_s
            end
          }.call
        rescue SecurityError
          @kernel_to_s.bind(obj).call
        end
      end
    }

    ###########################

    def self.<<(obj)
      if @debug_mode[1] # trust?
        @wrapper.call(obj.to_s)
      else
        proc{ $SAFE = 4; @wrapper.call(@inspector.call(obj, false)) }.call
      end
      self
    end

    def self.p(*args)
      if @debug_mode[1] # trust?
        args.each{|obj|
          @wrapper.call(obj.inspect + "\n")
        }
      else
        args.each{|obj|
          proc{ 
            $SAFE = 4
            @wrapper.call(@inspector.call(obj, true) + "\n")
          }.call
        }
      end
      nil
    end

    def self.putc(ch)
      m = @kernel_kind_of.bind(ch)

      if m.call(String)
        str = (@debug_mode[1])? ch[0,1]: proc{$SAFE=4; ch[0,1]}.call
      elsif m.call(Integer)
        str = (@debug_mode[1])? ch.chr: proc{$SAFE=4; ch.chr}.call
      else
        fail TypeError, 'cannot convert to Integer' 
      end

      @wrapper.call(str)
      ch
    end

    def self.puts(*args)
      if @debug_mode[1] # trust?
        args.each{|obj|
          if obj == nil
            @wrapper.call("nil\n")
          elsif @kernel_kind_of.bind(obj).call(String)
            obj = @inspector.call(obj, false)
            if obj =~ /\n\Z/
              @wrapper.call(obj)
            else
              @wrapper.call(obj + "\n")
            end
          elsif @kernel_kind_of.bind(obj).call(Array)
            self.puts(*obj)
          elsif obj.respond_to?(:to_ary) && 
              @kernel_kind_of.bind(ary = arg.to_ary).call(Array)
            self.puts(*ary)
          else
            str = @inspector.call(obj.to_s, false)
            unless @kernel_kind_of.bind(str).call(String)
              fail TypeError, 'cannot convert to String' 
            end
            if str =~ /\n\Z/
              @wrapper.call(str)
            else
              @wrapper.call(@inspector.call(str + "\n", false))
            end
          end
        }
      else
        args.each{|obj|
          if obj == nil
            @wrapper.call("nil\n")
          elsif @kernel_kind_of.bind(obj).call(String)
            obj = @inspector.call(obj, false)
            if obj =~ /\n\Z/
              @wrapper.call(obj)
            else
              @wrapper.call(@inspector.call(proc{$SAFE=4; obj + "\n"}.call, 
                                            false))
            end
          elsif @kernel_kind_of.bind(obj).call(Array)
            self.puts(*obj)
          elsif proc{
              $SAFE=4
              (obj.respond_to?(:to_ary) && 
                @kernel_kind_of.bind(ary = arg.to_ary).call(Array))
            }.call
            self.puts(*ary)
          else
            str = @inspector.call(proc{$SAFE=4; obj.to_s}.call, false)
            unless @kernel_kind_of.bind(str).call(String)
              fail TypeError, 'cannot convert to String' 
            end
            if proc{$SAFE=4; str =~ /\n\Z/}.call
              @wrapper.call(@inspector.call(str, false))
            else
              @wrapper.call(@inspector.call(proc{$SAFE=4; str + "\n"}.call))
            end
          end
        }
      end

      nil
    end

    def self.print(*args)
      if args.empty?
        self.print(@inspector.call($_, false))
      else
        if @debug_mode[1] # trust?
          str = args.collect{|obj|
            obj.to_s
          }.join(($, == nil)? '': @inspector.call($,, false))
        else
          str = args.collect{|obj|
            @inspector.call($,, false)
          }.join(($, == nil)? '': @inspector.call($,, false))
        end

        @wrapper.call(str)
      end
      nil
    end

    def self.printf(format, *args)
      if @debug_mode[1] # trust?
        str = @inspector.call(format, false) % args
      else
        str = proc{$SAFE=4; @inspector.call(format, false) % args}.call
      end
      self.print(str)
    end

    def self.write(str)
      if @debug_mode[1] # trust?
        new_str = str.to_s
      else
        begin
          new_str = str.to_s
        rescue
          new_str = @inspector.call(str, false)
        end
      end

      str = new_str
      unless @kernel_kind_of.bind(str).call(String)
        raise TypeError, 'cannot convert to String' 
      end
      @wrapper.call(str)
      @string_length.bind(str).call
    end

    def self.warn(msg)
      self.print(@inspector.call(msg, false), "\n") unless $VERBOSE.nil?
      nil
    end
  end

  DEBUG.freeze
end
