#
#  tkorca/method_wrapper.rb : function/singleton/instance method wrapper module
#
#                           by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#
if $0 == __FILE__
  $LOAD_PATH << File.expand_path(File.join(File.dirname(__FILE__), '..'))
end

require 'tkorca/utils'
require 'tkorca/debug'

module TkORCA
  module MethodWrapper
    @MethodWrapperProc = Class.new(Object){
      allows = ['class', '__id__', '__send__', 'object_id', 'to_s', 'freeze']
      undef_method(*(instance_methods - allows - allows.map{|s| s.to_sym}))
      protected :class

      class << self
        allows = ['new', '__id__', '__send__', 'object_id', 'freeze']
        undef_method(*(instance_methods - allows - allows.map{|s| s.to_sym}))

        def inspect
          id = __id__
          '#<Class:%#0*x>' % [id.size * 2 + 2, id]
        end
        alias to_s inspect
      end

      def initialize(singleton, method, body)
        @org_method = method

        if singleton
          # singleton method
          @proc = proc{|blk, *args|
            body.call(method, *args, &blk)
          }
        else
          # instance method
          if body.kind_of?(UnboundMethod)
            @proc = proc{|blk, obj, *args| 
              body.bind(obj).call(method.bind(obj), *args, &blk)
            }
          else
            @proc = proc{|blk, obj, *args| 
              body.call(method.bind(obj), *args, &blk)
            }
          end
        end

        self.freeze
      end

      def inspect
        id = __id__
        '#<%s:%#0*x>' % [self.class.inspect, id.size * 2 + 2, id]
      end
      alias to_s inspect

      def call(*args, &blk)
        @proc.call(blk, *args)
      end

      def org_method
        @org_method
      end
    }

    @MethodWrapperProc.freeze

    ###############################################################

    class UnknownWrappedMethodError < NoMethodError; end

    ###############################################################

    @Kernel = Kernel
    @singleton_method = {}
    @instance_method  = {}

    bt_remove_size = 3  # based on class/method depnedency in this file
    @modify_backtrace = proc{|e|
      bt = e.backtrace
      begin
        bt[-(caller.size)-1, bt_remove_size] = []
      rescue Exception
      end
      e.set_backtrace(bt)
    }

    class << self
      allows = ['__id__', '__send__', 'object_id', 'inspect', 'to_s', 'freeze', 'method']
      undef_method(*(instance_methods - allows - allows.map{|s| s.to_sym}))

      ################################
      private

      def _argument_check(obj, m, body, blk)
        TkORCA.security_check()

        m = m.to_sym if m.kind_of?(String)
        unless m.kind_of?(Symbol)
          raise ArgumentError, 'expect a Symbol or String as a method name'
        end

        body = blk unless body
        unless body
          raise ArgumentError, 'no method body'
        end

        [obj.__id__, m, body]
      end
    end

    ################################

    def self._call_with_denying_daughter(orig_m, *args, &blk)
      # TkORCA::DEBUG.p(["_call_with_denying_daughter $SAFE=#{$SAFE}", 
      #                   self, orig_m, args, blk].inspect)
      TkORCA.security_check()
      orig_m.call(*args, &blk)
    end

    @call_with_denying_daughter = method(:_call_with_denying_daughter)

    class << self
      undef_method('method')
      undef_method('_call_with_denying_daughter')
    end


    ######################################
    ##  for singleton methods
    ######################################
    def self._call_singleton_method(id, m, *args, &blk)
      # TkORCA::DEBUG.p($SAFE)  # caller's safe-level
      # TkORCA::DEBUG.p(['call_singleton', id, m, args, blk])
      if @singleton_method[id] && @singleton_method[id][m]
        begin
          @singleton_method[id][m].call(*args, &blk)
        rescue Exception => e
          @modify_backtrace.call(e)
          raise e
        end
      else
        raise UnknownWrappedMethodError
      end
    end

    def self.call_singleton_method(obj, m, *args, &blk)
      begin
        _call_singleton_method(obj.__id__, m, *args, &blk)
      rescue UnknownWrappedMethodError
        raise NoMethodError,"unknown wrapped method `#{m}' is called"
      end
    end

    def self.wrap_singleton_method(obj, m, body = nil, &blk)
      id, m, body = _argument_check(obj, m, body, blk)

      if @singleton_method[id] && @singleton_method[id][m]
        TkORCA::DEBUG.warn("Warning: already exist a wrapper for singleton method `#{m}'")
        method = @singleton_method[id][m].org_method
      else
        method = obj.method(m)
      end

      @singleton_method[id] ||= {}
      @singleton_method[id][m] = @MethodWrapperProc.new(true, method, body)

      obj.instance_eval("
        def #{m}(*args, &blk)
          #{self}._call_singleton_method(#{id}, :#{m}, *args, &blk)
        end
      ")

      nil
    end
    class << self
      alias wrap_module_method wrap_singleton_method
      alias wrap_class_method  wrap_singleton_method

      alias wrap_public_singleton_method wrap_singleton_method
      alias wrap_public_module_method    wrap_singleton_method
      alias wrap_public_class_method     wrap_singleton_method
    end

    def self.wrap_private_singleton_method(obj, m, body = nil, &blk)
      wrap_singleton_method(obj, m, body, &blk)
      obj.instance_eval("class << self; private :#{m}; end")
      nil
    end
    class << self
      alias wrap_private_module_method wrap_private_singleton_method
      alias wrap_private_class_method  wrap_private_singleton_method
    end

    def self.wrap_protected_singleton_method(obj, m, body = nil, &blk)
      wrap_singleton_method(obj, m, body, &blk)
      obj.instance_eval("class << self; protected :#{m}; end")
      nil
    end
    class << self
      alias wrap_protected_module_method wrap_protected_singleton_method
      alias wrap_protected_class_method  wrap_protected_singleton_method
    end

    def self.deny_daughter_to_singleton_methods(obj, *methods)
      methods.each{|m|
        wrap_singleton_method(obj, m, @call_with_denying_daughter)
      }
    end
    class << self
      alias deny_daughter_to_module_methods deny_daughter_to_singleton_methods
      alias deny_daughter_to_class_methods  deny_daughter_to_singleton_methods

      alias deny_daughter_to_public_singleton_methods \
              deny_daughter_to_singleton_methods
      alias deny_daughter_to_public_module_methods \
              deny_daughter_to_singleton_methods
      alias deny_daughter_to_public_class_methods \
              deny_daughter_to_singleton_methods
    end

    def self.deny_daughter_to_private_singleton_methods(obj, *methods)
      methods.each{|m|
        wrap_private_singleton_method(obj, m, @call_with_denying_daughter)
      }
    end
    class << self
      alias deny_daughter_to_private_module_methods \
              deny_daughter_to_private_singleton_methods
      alias deny_daughter_to_private_class_methods \
              deny_daughter_to_private_singleton_methods
    end

    def self.deny_daughter_to_protected_singleton_methods(obj, *methods)
      methods.each{|m|
        wrap_protected_singleton_method(obj, m, @call_with_denying_daughter)
      }
    end
    class << self
      alias deny_daughter_to_protected_module_methods \
              deny_daughter_to_protected_singleton_methods
      alias deny_daughter_to_protected_class_methods \
              deny_daughter_to_protected_singleton_methods
    end


    ######################################
    ##  for instance methods
    ######################################
    def self._call_instance_method(obj, m_klass_id, m, *args, &blk)
      # TkORCA::DEBUG.p($SAFE)  # caller's safe-level
      # TkORCA::DEBUG.p(['call_instance', obj, m_klass_id, m, args, blk])
      if @instance_method[m_klass_id] && @instance_method[m_klass_id][m]
        begin
          @instance_method[m_klass_id][m].call(obj, *args, &blk)
        rescue Exception => e
          @modify_backtrace.call(e)
          raise e
        end
      else
        raise UnknownWrappedMethodError
      end
    end

    def self.call_instance_method(obj, m_klass, m, *args, &blk)
      begin
        _call_instance_method(obj, m_klass.__id__, m, *args, &blk)
      rescue UnknownWrappedMethodError
        raise NoMethodError,"unknown wrapped method `#{m}' is called"
      end
    end

    def self.wrap_instance_method(klass, m, body = nil, &blk)
      id, m, body = _argument_check(klass, m, body, blk)

      if @instance_method[id] && @instance_method[id][m]
        TkORCA::DEBUG.warn("Warning: already exist a wrapper for instance method `#{m}'")
        method = @instance_method[id][m].org_method
      else
        method  = klass.instance_method(m)
      end

      @instance_method[id] ||= {}
      @instance_method[id][m] = @MethodWrapperProc.new(false, method, body)

      klass.module_eval("
        def #{m}(*args, &blk)
          #{self}._call_instance_method(self, #{id}, :#{m}, *args, &blk)
        end
      ")

      nil
    end
    class << self
      alias wrap_public_instance_method wrap_instance_method
    end

    def self.wrap_private_instance_method(klass, m, body = nil, &blk)
      wrap_instance_method(klass, m, body, &blk)
      klass.module_eval("private :#{m}")
      nil
    end

    def self.wrap_protected_instance_method(klass, m, body = nil, &blk)
      wrap_instance_method(klass, m, body, &blk)
      klass.module_eval("protected :#{m}")
      nil
    end

    def self.deny_daughter_to_instance_methods(klass, *methods)
      methods.each{|m|
        wrap_instance_method(klass, m, @call_with_denying_daughter)
      }
    end
    class << self
      alias deny_daughter_to_public_instance_methods \
              deny_daughter_to_instance_methods
    end

    def self.deny_daughter_to_private_instance_methods(klass, *methods)
      methods.each{|m|
        wrap_private_instance_method(klass, m, @call_with_denying_daughter)
      }
    end

    def self.deny_daughter_to_protected_instance_methods(klass, *methods)
      methods.each{|m|
        wrap_protected_instance_method(klass, m, @call_with_denying_daughter)
      }
    end

    ######################################
    ##  for functions
    ######################################
    def self.call_function(m, *args)
      begin
        _call_singleton_method(@Kernel.__id__, m, *args)
      rescue UnknownWrappedMethodError
        raise NoMethodError,"unknown wrapped function `#{m}' is called"
      end
    end

    def self.wrap_function(m, body = nil, &blk)
      wrap_singleton_method(@Kernel, m, body, &blk)
      wrap_private_instance_method(@Kernel, m, body, &blk)
      nil
    end

    def self.deny_daughter_to_functions(*methods)
      methods.each{|m|
        deny_daughter_to_singleton_methods(@Kernel, m)
        deny_daughter_to_private_instance_methods(@Kernel, m)
      }
      nil
    end
  end

  ##################################################
  module SINGLETON_METHODS
    def self.wrap(*args, &blk)
      MethodWrapper.wrap_singleton_method(*args, &blk)
    end
    class << self
      alias wrap_public wrap
    end
    def self.wrap_private(*args, &blk)
      MethodWrapper.wrap_private_singleton_method(*args, &blk)
    end
    def self.wrap_protected(*args, &blk)
      MethodWrapper.wrap_protected_singleton_method(*args, &blk)
    end

    def self.deny_daughter(*args)
      MethodWrapper.deny_daughter_to_singleton_methods(*args)
    end
    class << self
      alias deny_daughter_to_public deny_daughter
    end
    def self.deny_daughter_to_private(*args)
      MethodWrapper.deny_daughter_to_private_singleton_methods(*args)
    end
    def self.deny_daughter_to_protected(*args)
      MethodWrapper.deny_daughter_to_protected_singleton_methods(*args)
    end

    self.freeze
  end
  CLASS_METHODS  = SINGLETON_METHODS
  MODULE_METHODS = SINGLETON_METHODS

  module INSTNANCE_METHODS
    def self.wrap(*args, &blk)
      MethodWrapper.wrap_instance_method(*args, &blk)
    end
    class << self
      alias wrap_public wrap
    end
    def self.wrap_private(*args, &blk)
      MethodWrapper.wrap_private_instance_method(*args, &blk)
    end
    def self.wrap_protected(*args, &blk)
      MethodWrapper.wrap_protected_instance_method(*args, &blk)
    end

    def self.deny_daughter(*args)
      MethodWrapper.deny_daughter_to_instance_methods(*args)
    end
    class << self
      alias deny_daughter_to_public deny_daughter
    end
    def self.deny_daughter_to_private(*args)
      MethodWrapper.deny_daughter_to_private_instance_methods(*args)
    end
    def self.deny_daughter_to_protected(*args)
      MethodWrapper.deny_daughter_to_protected_instance_methods(*args)
    end

    self.freeze
  end

  module FUNCTIONS
    def self.wrap(*args, &blk)
      MethodWrapper.wrap_function(*args, &blk)
    end

    def self.deny_daughter(*args)
      MethodWrapper.deny_daughter_to_functions(*args)
    end

    self.freeze
  end
  ##################################################

  MethodWrapper.freeze
end

##########################################################################
# test
##########################################################################
if $0 == __FILE__
  TkORCA::DEBUG.debug_mode = true

  p Kernel.methods.find_all{|m| m =~ /req/}
  p Kernel.instance_methods.find_all{|m| m =~ /req/}

  begin
    Kernel.require('a')
  rescue LoadError 
  end
  begin
    require('a')
  rescue LoadError
  end

##########################################################################

#------------------------------------------------------------------
#  << definition-style 1 >>  ---  a block
#
#     This style cannot pass a block to the original method.
#     When ruby-1.9.x, a Proc object can accept a block. 
#     But, when ruby-1.8.x, cannot accept a block.
#     Therefore if you need to pass a block (e.g. 'each' method) 
#     on ruby-1.8.x, you have to use << definition-style 2 >> 
#------------------------------------------------------------------
TkORCA::MethodWrapper.wrap_function(:require){|orig_m, *args|
  p ["wrapped_function_call $SAFE=#{$SAFE}", self, orig_m, args]
  orig_m.call(*args)
}
# if ruby-1.9.x
#TkORCA::MethodWrapper.wrap_function(:require){|orig_m, *args, &blk|
#  p ["wrapped_function_call $SAFE=#{$SAFE}", self, orig_m, args, &blk]
#  orig_m.call(*args, &blk)
#}

#------------------------------------------------------------------
#  << definition-style 2 >>  ---  a method
#
#     This style can pass a block to the original method.
#     See also the following exsample (wrap methods of class A).
#------------------------------------------------------------------
#def wrapped_require(orig_m, file, &blk)
#  p ["wrapped_function_call $SAFE=#{$SAFE}", self, orig_m, file, blk]
#  orig_m.call(file, &blk)
#end
#TkORCA::MethodWrapper.wrap_function(:require, method(:wrapped_require))

##########################################################################

  p Kernel.methods.find_all{|m| m =~ /req/}
  p Kernel.instance_methods.find_all{|m| m =~ /req/}

  begin
    Kernel.require('a')
  rescue LoadError
  end
  begin
    require('a')
  rescue LoadError
  end
  begin
    p 'check $SAFE==4'
    proc{$SAFE=4; require('a')}.call
  rescue LoadError=>e
    #p e
  end

  class A
    def foo(a, b, c)
      p [:foo, a, b, c]
    end
    def each(i=0)
      yield(i+1)
      yield(i+2)
      yield(i+3)
    end
    def self.class_each(i=10)
      yield(i+1)
      yield(i+2)
      yield(i+3)
    end
  end

  class B < A
    def bar(a, b, c)
      p [:bar, a, b, c]
    end
  end

  A.new.foo(1,2,3)
  B.new.foo(1,2,3)

  A.new.each{|i| p i}
  A.new.each(2){|i| p i}

  TkORCA::MethodWrapper.wrap_instance_method(A, :foo){|orig_m, *args|
    p ["wrapped_method_call $SAFE=#{$SAFE}", self, orig_m, args]
    orig_m.call(*args)
  }
  #---------------------------------
  #def wrapped_foo(orig_m, a, b, c)
  #  p ["wrapped_method_call $SAFE=#{$SAFE}", self, orig_m, [a, b, c]]
  #  orig_m.call(a, b, c)
  #end
  #TkORCA::MethodWrapper.wrap_instance_method(A, :foo, method(:wrapped_foo))

  A.new.foo(1,2,3)
  B.new.foo(1,2,3)

  p '------------'

  b = B.new
  b.foo(1,2,3)
  b.bar(1,2,3)

  p '------------'

  TkORCA::MethodWrapper.deny_daughter_to_instance_method(B, :foo, :bar)
  b.foo(1,2,3)
  b.bar(1,2,3)

  p '------------'
  # Once wrapped, there is no effect of overriding the parrent method.

  class A
    def foo(a, b, c)
      p [:foo_2, a, b, c]
    end
  end

  A.new.foo(1,2,3)
  b.foo(1,2,3)
  B.new.foo(1,2,3)

  p '------------'

  #--------------------------------
  # To wrap A#each needs << definition-style 2 >> to pass a iterator block on 
  # ruby-1.8.x.
  # In this case, use a instance method to wrap the original method.
  #--------------------------------
  class A
    def wrapped_each(orig_m, *arg, &blk)
      p ["wrapped_method_call $SAFE=#{$SAFE}", self, orig_m, arg, blk]
      orig_m.call(*arg, &blk)
    end
    TkORCA::MethodWrapper.wrap_instance_method(self, :each, 
                                               instance_method(:wrapped_each))
  end
  #--------------------------------
  # The following is similar to the previous.
  # If use a instance method, it must belong to the same class.
  #--------------------------------
  #class A
  #  def wrapped_each(orig_m, *arg, &blk)
  #    p ["wrapped_method_call $SAFE=#{$SAFE}", self, orig_m, arg, blk]
  #    orig_m.call(*arg, &blk)
  #  end
  #end
  #TkORCA::MethodWrapper.wrap_instance_method(A, :each, 
  #                                           A.instance_method(:wrapped_each))
  #--------------------------------
  # In the following, use a class method
  #--------------------------------
  #class A
  #  def self.wrapped_each(orig_m, *arg, &blk)
  #    p ["wrapped_method_call $SAFE=#{$SAFE}", self, orig_m, arg, blk]
  #    orig_m.call(*arg, &blk)
  #  end
  #end
  #TkORCA::MethodWrapper.wrap_instance_method(A, :each, 
  #                                           A.method(:wrapped_each))

  A.new.each{|i| p i}
  A.new.each(2){|i| p i}

  p '------------'

  A.class_each{|i| p i}
  A.class_each(2){|i| p i}

  class A
    def self.wrapped_class_each(orig_m, *arg, &blk)
      p ["wrapped_method_call $SAFE=#{$SAFE}", self, orig_m, arg, blk]
      p "call TkORCA.security_check : ThreadGroup current=#{Thread.current.group}  Default=#{ThreadGroup::Default}"
      TkORCA.security_check()
      orig_m.call(*arg, &blk)
    end
    TkORCA::MethodWrapper.wrap_singleton_method(self, :class_each, 
                                                method(:wrapped_class_each))
#    TkORCA::MethodWrapper.deny_daughter_to_singleton_method(self, :class_each)
  end

  p '------------'

  A.class_each{|i| p i}
  A.class_each(2){|i| p i}

  p '------------'
  p "*** You'll get a `SecurityError' message."

  g = ThreadGroup.new # create non-default ThreadGroup ( == NOT mother )
  th = Thread.new{
    g.add(Thread.current)  # change ThreadGroup (simulate daughter process)
    A.class_each{|i| p i*5}
  }
  begin
    th.join
  rescue SecurityError=>e
    p e
  end

  p '------------'

  A.class_each{|i| p i*5}
end
