#
#  tkorca/mother.rb : Ruby/TkORCA Mother class
#
#                         by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#

module TkORCA
  class Mother
    DEFAULT_TIMEOUT = 3600 #second

    def inspect
      super() =~ /^(\S+)\s/
      $1 + '>'
    end

    ###############################

    def _signal_handler(sig)
      TkORCA.security_check(5)

      sig = sig.to_s
      case sig
      when 'INT', 'SIGINT'
        type = 'interrupt'
      when 'TERM', 'SIGTERM'
        type = 'terminate'
      when 'HUP', 'SIGHUP'
        type = 'hangup'
      else
        # not support
        type = ''
      end

      @close_notify_msg = 
        "Sorry. This connection received a #{type} signal.\n" + 
        'It will be down soon.'

      @msg_timeout = 5
      reset_timeout_timer(0)
    end

    def _exit_mother
      TkORCA.security_check(5)

      # force exit
      #Thread.new{sleep(3); TkORCA::EVLOOP_THREAD.kill; exit(0)}
      force_exit_th = Thread.new{
        Thread.pass
        sleep(1)
        TkORCA::EVLOOP_THREAD.kill
        Thread.pass
        sleep(1)
        # fail to exit -> force exit
        Kernel.exit!(0)
      }

      # delete daughters
      @daughter_table.each{|daughter|
        begin
          if daughter.alive?
            daughter._invoke('destroy', '.') rescue nil
            if daughter.safe?
              MultiTkIp._invoke_without_enc('::safe::interpDelete', 
                                            daughter.ip_name) rescue nil
            else
              MultiTkIp._invoke_without_enc('interp', 'delete', 
                                            daughter.ip_name) rescue nil
            end
          end
        rescue Exception
        end
      }

      # exit mother -- start exit operation on the other thread
      exit_th = Thread.new{ Tk.exit }

      # try to wakeup waiting threads
      10.times{|i|
        Thread.pass
        list = Thread.list - 
                [Thread.main, Thread.current, exit_th, force_exit_th]
        break if list.empty?
        list.each{|th| th.wakeup if th.stop?}
      }
    end

    def _remove_daughter(daughter)
      TkORCA.security_check(5)

      if @daughter_table.delete(daughter)
        daughter._eval('foreach id [after info] {after cancel $id}') rescue nil
        if daughter.alive?
          daughter._invoke('destroy', '.') rescue nil
          if daughter.safe?
            MultiTkIp._invoke_without_enc('::safe::interpDelete', 
                                          daughter.ip_name) rescue nil
          else
            MultiTkIp._invoke_without_enc('interp', 'delete', 
                                          daughter.ip_name) rescue nil
          end
        end
      end

      if @daughter_table.empty?
        _exit_mother
      end
    end

    def initialize(max_width = -1, max_height = -1)
      @timeout = DEFAULT_TIMEOUT
      @msg_timeout = 12

      @next_daughter = 0
      @daughter_table = []

      @wall = TkORCA::RFB_Canvas.new
      # @wall = TkORCA::RFB_Canvas.new(:wallbackground=>'skyblue')
      # @wall = RFB_Canvas.new(:wallbackground=>'#e0d6ff')
      # @wall = RFB_Canvas.new(:wallbackground=>'#d0d6cc')

      #Tk.update

      @max_width  = max_width
      @max_height = max_height

      @wall_width, @wall_height = @wall.get_wallsize

      @screen_width  = Tk.root.winfo_screenwidth
      @screen_height = Tk.root.winfo_screenheight

      if @max_width > 0 && @screen_width  > @max_width
        @screen_width  = @max_width
      end

      if @max_height > 0 && @screen_height > @max_height
        @screen_height = @max_height
      end

      @wall_width  = @screen_width  if @screen_width  > @wall_width
      @wall_height = @screen_height if @screen_height > @wall_height

      @wall.set_wallsize(@wall_width, @wall_height)

      #Tk.root.geometry("#{@screen_width}x#{@screen_height}-0-0")
      Tk.root.geometry("#{@screen_width}x#{@screen_height}+0+0")

      @close_notify_msg = 'Sorry. Your connection time is up.'
      @timeout_proc = proc{
=begin
        if ip
          rndname = '.___grab_dummy___'
          rndname = '.'
          ip.eval_proc{
            while TkWinfo.exist?(rndname)
              rndname << (?a + rand(26)).chr
            end
          }
          MultiTkIp.invoke_hidden(ip, 'toplevel', rndname)
          MultiTkIp.invoke_hidden(ip, 'wm', 'withdraw', rndname)
          MultiTkIp.invoke_hidden(ip, 'grab', rndname)
        end
=end
        #after_id = Tk.after(@msg_timeout * 1000){ Tk.root.destroy; exit }
        after_id = Tk.after((@msg_timeout * 1000).to_int){
          # 'messageBox' calls vwait for the variable '::tk::Priv(button)'
          Tk.ip_eval('set ::tk::Priv(button) ok')
        }
        Tk.messageBox(:type=>'ok', :message=>@close_notify_msg)
        Tk.after_cancel(after_id)
        # Tk.root.destroy
        # Thread.new{Tk.root.destroy}
        Thread.new{_exit_mother()}
      }

      @timeout_timer_id = nil
      if @timeout.kind_of?(Numeric) && @timeout > 0
        reset_timeout_timer(@timeout)
      end
    end

    def timeout
      @timeout
    end

    def reset_timeout_timer(timeout = nil)
      TkORCA.security_check(5)

      if timeout
        fail TypeError, 'expect Numeric' unless timeout.kind_of?(Numeric)
        fail ArgumentError, 'unexpect negative number' unless timeout >= 0
        @timeout = timeout
      end
      @timeout = DEFAULT_TIMEOUT unless @timeout

      Tk.after_cancel(@timeout_timer_id) if @timeout_timer_id

      @timeout_timer_id = Tk.after((@timeout * 1000).to_int, @timeout_proc)

      @timeout
    end

    def wall
      @wall
    end

    def daughters
      TkORCA.security_check(5)
      @daughter_table.dup
    end

    def new_daughter(trust=false, safe_level=4, init_proc=nil, wall=nil, &b)
      wall = @wall unless wall

      # w = TkORCA::Window_Frame.new(wall, 'slave root', 50, 70, :anchor=>:nw)
      # w = TkORCA::Window_Frame.new(wall, 'slave root', 50, 70)
      w = TkORCA::Window_Frame.new(wall, 'slave root')

      # p '---------------------'
      # p MultiTkIp.__getip
      # p w.id
      # p wall.find_withtag(TkcNamedTag.new(wall, 'window_frame_tag')).collect{|item| item.inspect[0..30]}
      # p '---------------------'

      #ip = MultiTkIp.new(:use=>w.winid)
      #ip = MultiTkIp.new_trusted_slave(:use=>w.winid)
      #ip = MultiTkIp.new_trusted_slave(3, :use=>w.winid)
      #ip = MultiTkIp.new_safeTk(0, :use=>w.winid)
      #ip = MultiTkIp.new_safeTk(3, :use=>w.winid)
      #ip = MultiTkIp.new_safeTk(:use=>w.winid)

      # p '---------------------'
      # p MultiTkIp.__getip
      # p w.id
      # p wall.find_withtag(TkcNamedTag.new(wall, 'window_frame_tag')).collect{|item| item.inspect[0..30]}
      # p '---------------------'

      @daughter_table << (ip = Daughter.new(trust, safe_level, w.winid, 
                                            TkWinfo.screen(wall.path)))

      # w.instance_variable_get(:@container).bind_append('Destroy'){Tk.exit}
      w.instance_variable_get(:@container).bind_append('Destroy'){
        _remove_daughter(ip)
      }

      # p '---------------------'
      # p MultiTkIp.__getip
      # p w.id
      # p wall.find_withtag(TkcNamedTag.new(wall, 'window_frame_tag')).collect{|item| [item, wall.itemcget(item, :window)]}
      # p '---------------------'

      TkORCA::TclCmdWrapper.apply_wrapper_modules(ip)

      init_proc = b if !init_proc && b
      init_proc.call(ip) if init_proc

      ip
    end
  end
end
