#
#  tkorca/parse_app_opt.rb : utility method to parse option about application
#
#      app_info = app_path,trusted_ip,safe_level,use_console,init_script
#
#            When each param is a empty string, use default for the param.
#            If init_script is "{ ... }" (wrapped by curly brace), it means 
#            a ruby script. Else, means a filename of the initialize script.
#
#                           by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#
module TkORCA
  module Parse_Application_Opt
    @init_proc_available = true
  end
end

class << TkORCA::Parse_Application_Opt
  def _file_value(file)
    file = File.join(SCRIPT_DIR, file) unless file =~ %r{^[/~]}
    File.expand_path(file)
  end

  def _trust_value(trust)
    trust = TkComm.bool(trust) if trust
    (trust.nil?)? false: trust
  end

  def _safe_value(safe)
    (safe)? safe.to_i: 4
  end

  def _console_value(console)
    console = TkComm.bool(console) if console
    (console.nil?)? false: console
  end

  def _init_value(init)
    return nil unless init && @init_proc_available

    init = init.strip
    if init =~ /\A[{].*[}]\Z/m
      # is script
      init_script = "proc#{init}.call"
    else
      # is filename
      init_script = "load('#{_file_value(init)}')"
    end
    (init_script)? proc{|ip| eval(init_script)}: nil
  end
  private :_trust_value, :_safe_value, :_console_value, :_init_value

  def get_default_info(table, params)
    table.each{|info|
      next if params['file'] != info[0] && params['selected_file'] != info[0]

      next if params.key?('trusted_slave') && 
        ((info[2])? !params['trusted_slave']: params['trusted_slave'])

      next if params.key?('safe_level') && params['safe_level'] != info[3]

      next if params.key?('use_console') && 
        ((info[4])? !params['use_console']: params['use_console'])

      next if params.key?('init_proc')

      return info
    }
    nil
  end

  def parse_app_opt(app_info, default_table=nil)
    file, trust, safe, console, init = 
      app_info.split(',', 5).collect{|s| s.strip}

    trust   = nil if trust   && trust.empty?
    safe    = nil if safe    && safe.empty?
    console = nil if console && console.empty?
    init    = nil if init    && init.empty?

    selected_file = _file_value(file)
    trusted_slave = _trust_value(trust)
    safe_level    = _safe_value(safe)
    use_console   = _console_value(console)
    init_proc     = _init_value(init)

    check_params = {'file' => file, 'selected_file' => selected_file}
    check_params['trusted_slave'] = trusted_slave if trust
    check_params['safe_level']    = safe_level    if safe
    check_params['use_console']   = use_console   if console
    check_params['init_proc']     = init_proc     if init

    # get defaults from table
    if default_table.kind_of?(Array)
      info = get_default_info(default_table, check_params)
      if info
        trusted_slave = info[2] unless trust
        safe_level    = info[3] unless safe
        use_console   = info[4] unless console
        init_proc     = info[5] unless init
      end
    end

    [selected_file, trusted_slave, safe_level, use_console, init_proc]
  end

  def disable_init_proc
    @init_proc_available = false
    self.freeze
  end
end
