#
#  rfb_canvas.rb : `wall' class of Ruby/TkORCA
#
#                           by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#

require 'tkorca/utils'

module TkORCA
  class RFB_Canvas < TkCanvas
    include TkComposite

    @@broadband_network = false

#    DEFAULT_WIDTH  = 640
#    DEFAULT_HEIGHT = 480
#    DEFAULT_WIDTH  = 1280
#    DEFAULT_HEIGHT = 1024
    DEFAULT_WIDTH  = 1600
    DEFAULT_HEIGHT = 1200

    ############################

    def inspect
      super() =~ /^(\S+)\s/
      $1 + '>'
    end

    ###############################

    @wall_table = {}
    @wall_tktbl = TkVariable.new_hash

    def self.new(keys = {}, &blk)
      keys = _symbolkey2str(keys)

      screen = keys.delete('screen').to_s

      if @wall_table[screen]
        @wall_table[screen]
      else
        root_scrn = TkWinfo.screen('.')
        if screen == '' || screen == root_scrn
          @wall_table[''] = @wall_table[root_scrn] = super(Tk.root, keys, &blk)
          @wall_tktbl[''] = @wall_tktbl[root_scrn] = @wall_table['']
        else
          @wall_table[screen] = super(TkToplevel.new(:screen=>screen),
                                      keys, &blk)
          @wall_tktbl[screen] = @wall_table[screen]
        end
      end
    end

    def self.get_wall_for_screen(screen)
      unless TkORCA.is_mother?
        raise SecurityError, 'available on TkORCA mother thread only'
      end

      @wall_table[screen.to_s]
    end

    def self.get_wallpath_for_screen(screen)
      wall = self.get_wall_for_screen(screen)
      if wall.kind_of?(TkWindow)
        wall.path
      else
        wall.to_s
      end
    end

    def self.get_wall
      self.get_wall_for_screen('')
    end

    def self.get_wallpath_table_id
      '::' << @wall_tktbl.id
    end

    ############################

    def initialize(toplevel, keys)
      @toplevel = toplevel

      if TkORCA.respond_to?(:test_screen_width)
        # local test mode
        @screen_width  = TkORCA.test_screen_width
      else
        @screen_width  = TkWinfo.screenwidth(@toplevel)
      end

      if TkORCA.respond_to?(:test_screen_height)
        # local test mode
        @screen_height = TkORCA.test_screen_height
      else
        @screen_height = TkWinfo.screenheight(@toplevel)
      end

      @toplevel.geometry("#{@screen_width}x#{@screen_height}-0-0")

      super(@toplevel, keys)
    end

    def initialize_composite(keys = {})
      @wall = TkCanvas.new(@frame)
      @path = @wall.path

      @wallpapers = []

      _set_canvas_bindings

      @broadband_network = @@broadband_network

      @h_scroll = TkScrollbar.new(@frame, :width=>10, 
                                  :jump=>!@broadband_network)
      @v_scroll = TkScrollbar.new(@frame, :width=>10, 
                                  :jump=>!@broadband_network)

      @wall.xscrollbar(@h_scroll)
      @wall.yscrollbar(@v_scroll)

      TkGrid.rowconfigure(@frame, 0, 'weight'=>1, 'minsize'=>0)
      TkGrid.rowconfigure(@frame, 1, 'weight'=>0)
      TkGrid.columnconfigure(@frame, 0, 'weight'=>1, 'minsize'=>0)
      TkGrid.columnconfigure(@frame, 1, 'weight'=>0)

      @wall.grid(:row=>0, :column=>0, :sticky=>'news')
      @h_scroll.grid(:row=>1, :column=>0, 'sticky'=>'ew')
      @v_scroll.grid(:row=>0, :column=>1, 'sticky'=>'ns')

      @frame.pack(:fill=>:both, :expand=>true)

      # set receiver widgets for configure methods (with alias)
      delegate_alias('wallbackground', 'background', @wall)
      delegate_alias('wallwidth',  'width', @wall)
      delegate_alias('wallheight', 'height', @wall)

      # set receiver widgets for configure methods
      delegate('DEFAULT', @wall)
      delegate('width', @frame)
      delegate('height', @frame)
      delegate('background', @h_scroll, @v_scroll)
      delegate('activebackground', @h_scroll, @v_scroll)
      delegate('troughcolor', @h_scroll, @v_scroll)
      delegate('repeatdelay', @h_scroll, @v_scroll)
      delegate('repeatinterval', @h_scroll, @v_scroll)
      delegate('borderwidth', @wall)
      delegate('relief', @wall)

      # do configure
      configure keys unless keys.empty?
    end

    def _set_canvas_bindings
      @wall.scrollregion = [0, 0, DEFAULT_WIDTH - 1, DEFAULT_HEIGHT - 1]
      @wall.bind('1', proc{|x,y| @wall.scan_mark x,y}, '%x %y')
      @wall.bind('B1-Motion', proc{|x,y| @wall.scan_dragto x,y}, '%x %y')
    end

    private :initialize_composite, :_set_canvas_bindings

    def broadband_mode?
      @broadband_network
    end
    def broadband_mode=(mode)
      mode = (mode)? true: false
      @broadband_network = mode
      @h_scroll.jump = !@broadband_network
      @v_scroll.jump = !@broadband_network
      mode
    end

    def get_wallsize
      x1, y1, x2, y2 = @wall.scrollregion
      [ x2 - x1 + 1, y2 - y1 + 1 ]
    end

    def set_wallsize(x, y)
      @wall.scrollregion = [0, 0, x - 1, y - 1]
      self
    end

    def put_wallpaper(img)
      unless img.kind_of?(TkImage)
        raise TypeError, 'expect a TkImage object'
      end

      @wall.delete(*@wallpapers)
      @wallpapers.clear

      dx = img.width
      dy = img.height

      x_min, y_min, x_max, y_max = self.scrollregion 

      while y_min <= y_max
        x = x_min
        while x <= x_max
          @wallpapers << TkcImage.new(@wall, x, y_min, 
                                      :anchor=>:nw, :image=>img).lower
          x += dx
        end
        y_min += dy
      end

      self
    end

    def write_msg(x, y, keys)
      TkcText.new(@wall, x, y, keys).raise(@wallpapers[0])
    end

    #####################
    def winfo_screenwidth
      self.winfo_width
    end

    def winfo_screenheight
      self.winfo_height
    end

    #####################
    def wall_vrootx
      self.canvasx(0).to_int
    end
    def wall_vrooty
      self.canvasy(0).to_int
    end

=begin
    def wall_x(root_x)
      self.canvasx(root_x - self.winfo_rootx).to_int
    end

    def wall_rooty(root_y)
      self.canvasy(root_y - self.winfo_rooty).to_int
    end
=end
#=begin
    def wall_rootx(root_x)
      (root_x - self.winfo_rootx).to_int
    end

    def wall_rooty(root_y)
      (root_y - self.winfo_rooty).to_int
    end
#=end
  end
end
