#
#  tclcmd_wrapper.rb : Tcl/Tk command wrapper module
#
#                           by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#
require 'thread'
require 'multi-tk'

if $0 == __FILE__
  $LOAD_PATH << File.expand_path(File.join(File.dirname(__FILE__), '..'))
end

require 'tkorca/utils'
require 'tkorca/debug'


module TkORCA
  module TclCmdWrapper
    def self.get_wall(daughter, win)
      screen = daughter._invoke('winfo', 'screen', win)
      TkORCA::RFB_Canvas.get_wall_for_screen(screen)
    end

    def self.get_container(daughter, win)
      top = daughter._invoke('winfo', 'toplevel', win)
      begin
        return nil if (use = daughter._invoke(top, 'cget', '-use')).empty?
      rescue
        # no '-use' option (maybe menu widget and so on)
        return nil
      end
      # screen = daughter._invoke(top, 'cget', '-screen')
      screen = daughter._invoke('winfo', 'screen', top)
      wall = TkORCA::RFB_Canvas.get_wall_for_screen(screen)
      wall.winfo_widget(use)
    end

    def self.get_window_frame(daughter, win)
      return nil unless (container = get_container(daughter, win))
      TkORCA::Window_Frame.get_frame(container)
    end

    ##########################

    class ReturnValue < StandardError
      def initialize(val = nil)
        super('succeed')
        @value = val
      end

      attr_reader :value
    end
    ReturnValue.freeze

    ##########################

    @wrapper_modules = []

    def self.entry_wrapper_module(mod)
      @wrapper_modules << mod
      mod
    end
    def self.apply_wrapper_modules(ip)
      @wrapper_modules.each{|mod|
        mod.replace_in(ip)
      }
      ip
    end

    ##########################

    @tclcmd_table = {}

    @cmd_queue = Queue.new
    @check_queue = Queue.new

    receiver_cmd = proc{
      loop {
        Thread.new(@cmd_queue.deq){|params|
          th, ip, cmd, args = params
          begin
            #xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            #  NOTE::  *** NEVER CALL LOGGING HERE ***
            #         It may break the callback sequence, 
            #         and make troubles on menu operations.
            #xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            # TkORCA::DEBUG.p(['tclcmd_wrapper', th, ip, cmd, args])
            #xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            if @tclcmd_table[cmd]
              ret = @tclcmd_table[cmd].call(ip, cmd, *args)
            else
              raise RuntimeError, "invalid command name `#{cmd}'"
            end
          rescue Exception=>e
            th.raise(e) if th
          else
            val = ReturnValue.new(ret)
            th.raise(ReturnValue.new(ret)) if th
          end
        }
      }
    }

    @receiver = [ Thread.new(&receiver_cmd) ]
    @watch_dog = Thread.new{
      loop{
        begin
          @receiver[0].join
        rescue Exception=>e
        end
        @receiver[0] = Thread.new(&receiver_cmd)
      }
    }

    def self.call(cmd, *args)
      begin
        th = Thread.new{ Thread.stop }
        Thread.pass
        @cmd_queue.enq([th, MultiTkIp.__getip, cmd, args])
        while th.alive?
          Tk.update
          Thread.pass
        end
        th.join
      rescue ReturnValue=>ret
        ret.value
      rescue Exception=>e
        TkORCA::DEBUG.p [e, e.backtrace]
        raise e
      else
        nil
      end
    end

    def self.bg_call(cmd, *args)
      @cmd_queue.enq([nil, MultiTkIp.__getip, cmd, args])
    end

    def self.wrap_tcl_command(slave_ip, cmd, body = nil, &blk)
      TkORCA.security_check()

      cmd = cmd.to_s unless cmd.kind_of?(String)

      body = blk unless body
      unless body
        if @tclcmd_table[cmd]
          body = @tclcmd_table[cmd]
        else
          raise ArgumentError, "no wrapper is given"
        end
      else
        @tclcmd_table[cmd] = body
      end

      # proc for calling the original cmd
      org_cmd = proc{|*args| MultiTkIp.invoke_hidden(slave_ip, cmd, *args)}

      # create callback
      #    * keep current $SAFE level
      #    * each of arguments given by Tcl/Tk is a UTF-8 string
      cb_proc = proc{|*args|
        args.each{|s| s.instance_variable_set('@encoding', 'utf-8')}
        self.call(cmd, *args)
      }
      cb_cmd = slave_ip.eval{TkComm._get_eval_string(cb_proc)}
      cb_cmd = cb_cmd.value if cb_cmd.kind_of?(Thread)

      # store original command
      unless MultiTkIp.hidden_cmds(slave_ip).index(cmd)
        MultiTkIp.hide_cmd(slave_ip, cmd)
      end

      # define wrapped command
      slave_ip._eval("proc #{cmd} args {eval \"#{cb_cmd} $args\"}")

      [slave_ip, cmd, org_cmd, body]
    end

    def self.wrap_tcl_command_transaction(slave_ip, cmd, transaction, 
                                          pre_cb = nil, post_cb = nil)
      TkORCA.security_check()

      cmd = cmd.to_s unless cmd.kind_of?(String)
      transaction = transaction.to_s unless transaction.kind_of?(String)

      if pre_cb
        # each of arguments given by Tcl/Tk is a UTF-8 string
        pre_cb_proc = proc{|*args| 
          args.each{|s| s.instance_variable_set('@encoding', 'utf-8')}
          TclTkLib._merge_tklist(*(self.call(cmd, *args)))
        }
        pre_cb_cmd = TkComm._get_eval_string(pre_cb_proc)  # entry on mother
      else
        pre_cb_cmd = ''
      end
      if post_cb
        # each of arguments given by Tcl/Tk is a UTF-8 string
        post_cb_proc = proc{|*args| 
          args.each{|s| s.instance_variable_set('@encoding', 'utf-8')}
          TclTkLib._merge_tklist(*(self.call(cmd, *args)))
        }
        post_cb_cmd = TkComm._get_eval_string(post_cb_proc)  # entry on mother
        post_cb_cmd << ' $transaction_result'
      else
        post_cb_cmd = ''
      end

      # store original command
      unless MultiTkIp.hidden_cmds(slave_ip).index(cmd)
        MultiTkIp.hide_cmd(slave_ip, cmd)
      end

      # define wrapped command
      MultiTkIp._eval <<-EOS
      proc slave_#{cmd} {slave args} {
        set pre_cb_result [#{pre_cb_cmd}]
        set transaction_result [#{transaction}]
        #{post_cb_cmd}
      }
      EOS
      MultiTkIp.def_alias(slave_ip.ip_name, cmd, 
                          'slave_' + cmd, slave_ip.ip_name)
    end

    def self.convert_tcl_command_args(slave_ip, cmd, body = nil, &blk)
      TkORCA.security_check()

      cmd = cmd.to_s unless cmd.kind_of?(String)

      body = blk unless body
      unless body
        if @tclcmd_table[cmd]
          body = @tclcmd_table[cmd]
        else
          raise ArgumentError, "no argument-checker is given"
        end
      else
        @tclcmd_table[cmd] = body
      end

      if Tk.info('command', "slave_#{cmd}").empty?
        # create callback
        #    * keep current $SAFE level
        #    * each of arguments given by Tcl/Tk is a UTF-8 string
        cb_proc = proc{|*args| 
          args.each{|s| s.instance_variable_set('@encoding', 'utf-8')}
          TclTkLib._merge_tklist(*(self.call(cmd,*args)))
        }
        cb_cmd = TkComm._get_eval_string(cb_proc)  # entry on mother

        # define wrapped command
        MultiTkIp._eval <<-EOS
        proc slave_#{cmd} {slave args} {
          eval interp invokehidden $slave #{cmd} [eval #{cb_cmd} $args]
        }
        EOS
      end

      # store original command
      unless MultiTkIp.hidden_cmds(slave_ip).index(cmd)
        MultiTkIp.hide_cmd(slave_ip, cmd)
      end

      # define alias on slave
      MultiTkIp.def_alias(slave_ip.ip_name, cmd, 
                          'slave_' + cmd, slave_ip.ip_name)
    end
  end
end

#########################

if $0 == __FILE__
  require 'tkorca/rfb_canvas.rb'
  require 'tkorca/get_wmframe.rb'
  require 'tkorca/window_frame.rb'

  TkORCA::DEBUG.debug_mode = true

  wall = TkORCA::RFB_Canvas.new
  w = TkORCA::Window_Frame.new(wall, 'slave root', 50, 70, :anchor=>:nw)
  ip = MultiTkIp.new_slave(4, :use=>w.winid)
  TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'toplevel'){|arg_ip ,cmd, *args|
    arg_ip._invoke('info', 'command', 'winfo')
    TkORCA::TclCmdWrapper.get_container(arg_ip, '.')
    [$SAFE, arg_ip.safe_level, arg_ip, cmd, args]
  }

  ret = ip.eval_proc{
    TkORCA::DEBUG.p $SAFE
    Tk.tk_call('toplevel', '.top')
  }
  raise ret if ret.kind_of?(Exception)
  p ret

  Tk.mainloop
end
