#
#  wrap_misc_tclcmd.rb : wrap misc Tcl command
#
#         *****  See "man n interp"  *****
#
#                           by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#
require 'tkorca/tclcmd_wrapper'

class << TkORCA::TclCmdWrapper.entry_wrapper_module(Module.new)
  def replace_in(ip)
    ########################################
    #   Based on "man n interp"
    ########################################
    # A  safe  interpreter is created with exactly the following
    # set of built-in commands:
    #
    #      after       append      array       binary
    #      break       case        catch       clock
    #      close       concat      continue    eof
    #      error       eval        expr        fblocked
    #      fcopy       fileevent   flush       for
    #      foreach     format      gets        global
    #      if          incr        info        interp
    #      join        lappend     lindex      linsert
    #      list        llength     lrange      lreplace
    #      lsearch     lsort       namespace   package
    #      pid         proc        puts        read
    #      regexp      regsub      rename      return
    #      scan        seek        set         split
    #      string      subst       switch      tell
    #      time        trace       unset       update
    #      uplevel     upvar       variable    vwait
    #      while
    #
    # The following commands are hidden by interp create when it
    # creates a safe interpreter:
    #
    #      cd          encoding    exec        exit
    #      fconfigure file         glob        load
    #      open        pwd         socket      source
    #
    # These commands can be recreated later as Tcl procedures or
    # aliases, or re-exposed by interp expose.
    #
    # The following commands from Tcl's library of support  pro-
    # cedures are not present in a safe interpreter:
    #
    #      auto_exec_ok    auto_import     auto_load
    #      auto_load_index auto_qualify    unknown
    #
    # Note  in particular that safe interpreters have no default
    # unknown command, so Tcl's default  autoloading  facilities
    # are not available.  Autoload access to Tcl's commands that
    # are normally autoloaded:
    #
    #      auto_mkindex         auto_mkindex_old
    #      auto_reset           history
    #      parray               pkg_mkIndex
    #      ::pkg::create        ::safe::interpAddToAccessPath
    #      ::safe::interpCreate ::safe::interpConfigure
    #      ::safe::interpDelete ::safe::interpFindInAccessPath
    #      ::safe::interpInit   ::safe::setLogCmd
    #      tcl_endOfWord        tcl_findLibrary
    #      tcl_startOfNextWord  tcl_startOfPreviousWord
    #      tcl_wordBreakAfter   tcl_wordBreakBefore
    #
    # can only be provided by explicit definition of an  unknown
    # command in the safe interpreter.  This will involve expos-
    # ing the source command.  This is most easily  accomplished
    # by creating the safe interpreter with Tcl's Safe-Tcl mech-
    # anism.  Safe-Tcl provides safe versions of  source,  load,
    # and  other  Tcl  commands needed to support autoloading of
    # commands and the loading of packages.

    TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'exec', &(_create_nop_proc()))

    TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'socket', &(_create_nop_proc()))

    _expose_cmd(ip, 'pwd')
    #TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'pwd', &(_create_nop_proc()))

    _expose_cmd(ip, 'fconfigure')
    #TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'fconfigure', &(_create_nop_proc()))

    _expose_cmd(ip, 'cd')
    # TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'cd', &(_create_nop_proc()))
    TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'cd', &(_create_cd_check_proc()))

    _expose_cmd(ip, 'glob')
    # TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'glob', &(_create_nop_proc()))
    TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'glob', &(_create_glob_check_proc()))

    # _expose_cmd(ip, 'open')
    # TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'open', &(_create_nop_proc()))
    TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'open', &(_create_open_check_proc()))

###########

    #TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'file'){|ip, cmd, *args|
    #  TkORCA::DEBUG.p [ip, cmd, *args]
    #  Tk.tk_call(cmd, *args)      
    #}

    # The followings are provided in a safe interpreter.
    # TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'file', &(_create_nop_proc()))
    # TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'load', &(_create_nop_proc()))
    # TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'source', &(_create_nop_proc()))
    # TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'encoding', &(_create_nop_proc()))
    # TkORCA::TclCmdWrapper.wrap_tcl_command(ip, 'exit', &(_create_nop_proc()))

    # `file' command on a safe interpreter accepts the following pattern of 
    # subcommands (See `AliasSubset' on <tcl-lib>/safe.tcl).
    # ------------------------------------------------------------------------
    #  AliasSubset $slave file file dir.* join root.* ext.* tail path.* split
    #                               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  end

  ######################################
  private
  ######################################
  def _create_nop_proc
    proc{|ip, cmd, *args|
      '' # return an empty string
    }
  end

  def _create_open_check_proc
    proc{|ip, cmd, *args|
      TkORCA::DEBUG.p [ip, cmd, *args]
      file = args.shift
      unless filepath = TkORCA::FileAccess.allow_tcl_to_open?(file, *args)
        fail LoadError, "cannot open file -- `#{file}'"
      end
      TkORCA::LOG.log(TkORCA::Daughter._name(ip), 
                      "[Tcl/Tk] open file `#{file}'")
      MultiTkIp.invoke_hidden(ip, cmd, file, *args)
    }
  end

  def _create_cd_check_proc
    proc{|ip, cmd, *args|
      TkORCA::DEBUG.p [ip, cmd, *args]
      dirname = args.shift
      if dirname.nil?
        MultiTkIp.invoke_hidden(ip, cmd)
      elsif dirpath = TkORCA::FileAccess.allow_tcl_to_cd?(dirname)
        MultiTkIp.invoke_hidden(ip, cmd, dirpath)
      else
        fail LoadError, "cannot change directory to `#{dirname}'"
      end
    }
  end

  def _create_glob_check_proc
    proc{|ip, cmd, *args|
      TkORCA::DEBUG.p [ip, cmd, *args]
      list = MultiTkIp.invoke_hidden(ip, cmd, *args)
      list = TkORCA::FileAccess.check_glob_result?(list, *args)
    }
  end

  def _expose_cmd(ip, cmd)
    if MultiTkIp.hidden_cmds(ip).index(cmd)
      MultiTkIp.expose_cmd(ip, cmd)
    end
  end
end
