#!/usr/local/bin/ruby
#
#  Ruby/TkORCA server script
#
#    tkorca_server.rb [ ...server options... ] 
#                  [-- [ ...tkorca options... ] [-- [ ...applications...]]]
#
#    If you want give applications only (no options), requrie two '--'.
#      e.g. tkorca_server.rb -- -- app1.rb app2.rb ...
#
#                             by Hidetoshi NAGAI (nagai@ai.kyutech.ac.jp)
#
require 'tmpdir'
require 'socket'

DEBUG = IO.open(2, 'w')


##############################################
#  load version information
##############################################
load File.join(File.dirname(__FILE__), 'version.rb')


########################################################
# add SERVER_DIR to load-path
########################################################
SERVER_DIR = File.dirname(File.expand_path(__FILE__))
$LOAD_PATH.unshift(SERVER_DIR)


########################################################
# setup base definitions
########################################################
load(File.join(SERVER_DIR, 'server_base_defs'))


########################################################
# define defaults
########################################################
unless Object.const_defined?(:TkORCA_SIGHUP)
  list = Signal.list
  if list.has_key?('HUP')
    TkORCA_SIGHUP = :HUP
  elsif list.has_key?('BREAK')
    TkORCA_SIGHUP = :BREAK
  else
    TkORCA_SIGHUP = nil
  end
end

unless Object.const_defined?(:NULL_DEVICE)
  NULL_DEVICE = '/dev/null'
end

unless Object.const_defined?(:USE_ID_CHANGE_PRIVILEGE)
  USE_ID_CHANGE_PRIVILEGE = true
end

unless Object.const_defined?(:TkORCA_SERVER_USER)
  # TkORCA_SERVER_USER = 'nobody'
  TkORCA_SERVER_USER = Process::UID.eid
end

unless Object.const_defined?(:TkORCA_SERVER_GROUP)
  # TkORCA_SERVER_GROUP = 'nobody'
  TkORCA_SERVER_GROUP = Process::GID.eid
end

unless Object.const_defined?(:TkORCA_USER)
  # TkORCA_USER = 'nobody'
  TkORCA_USER = Process::UID.eid
end

unless Object.const_defined?(:TkORCA_GROUP)
  # TkORCA_GROUP = 'nobody'
  TkORCA_GROUP = Process::GID.eid
end

unless Object.const_defined?(:HOST_ADDRESS)
  HOST_ADDRESS = ''
end

unless Object.const_defined?(:SERVICE_PORT)
  SERVICE_PORT = 5940
end

unless Object.const_defined?(:LISTEN_BACKLOG)
  LISTEN_BACKLOG = 5
end

unless Object.const_defined?(:TkORCA_INTERACTIVE)
  TkORCA_INTERACTIVE = true
end

#unless Object.const_defined?(:CONTROL_PORT)
#  CONTROL_PORT = SERVICE_PORT + 1
#end

unless Object.const_defined?(:REVERSE_LOOKUP)
  REVERSE_LOOKUP = false
end

unless Object.const_defined?(:SERVER_CONFDIR)
  # SERVER_CONFDIR = SERVER_DIR
  SERVER_CONFDIR = File.join(SERVER_DIR, 'sysconf')
end

unless Object.const_defined?(:SERVER_RCFILE)
  SERVER_RCFILE = File.join(SERVER_CONFDIR, 'tkorca_serverrc')
end

unless Object.const_defined?(:LOCAL_RCFILE)
  # LOCAL_RCFILE = '~/.tkorca_serverrc'
  LOCAL_RCFILE = nil
end

#unless Object.const_defined?(:SERVER_TMPDIR)
#  SERVER_TMPDIR = File.join(SERVER_DIR, 'tmp_dir')
#end

#unless Object.const_defined?(:SERVER_LOGFILE)
#  SERVER_LOGFILE = File.join(SERVER_TMPDIR, 'tkorca.log')
#  # SERVER_LOGFILE = '/var/log/tkorca.log'
#end

#unless Object.const_defined?(:SERVER_PID)
#  SERVER_PID = File.join(SERVER_TMPDIR, 'tkorca.pid')
#  # SERVER_PID = '/var/run/tkorca.pid'
#end

unless Object.const_defined?(:CONTROL_PASSWD_FILE)
  CONTROL_PASSWD_FILE = File.join(SERVER_CONFDIR, 'tkorca_passwd')
end

unless Object.const_defined?(:RFBAUTH_PASSWD_FILE)
  RFBAUTH_PASSWD_FILE = nil
end


########################################################
#  get command-line options
########################################################
require 'server/opt_parse.rb'

# rests of ARGV are options for TkORCA services.
# force parsing options for TkORCA services to get the list of applications.
if ARGV.empty?
  TkORCA_SERVICE_OPTIONS = []
else
  TkORCA_SERVICE_OPTIONS = ARGV.dup
  Module.new.module_eval(IO.read(File.join(SERVER_DIR, 'tkorca/opt_parse.rb')))
end

if ARGV.empty?
  TkORCA_SERVICE_APPLICATIONS = []
else
  TkORCA_SERVICE_OPTIONS[(-(ARGV.length))..(-1)] = []
  TkORCA_SERVICE_OPTIONS.pop if TkORCA_SERVICE_OPTIONS[-1] == '--'

  TkORCA_SERVICE_APPLICATIONS = ARGV.dup
end

=begin
# rests of ARGV are options for TkORCA services
# NOTE: expect [ ... options ... ] -- [ ... applications ... ] for rests.
#       The param `--' is important. However, if `fork_tkorca()' is called
#       without additional arguments, that will work properly without `--'.

TkORCA_SERVICE_OPTIONS = []
if ARGV[0] =~ /^-/
  until ARGV.empty?
    arg = ARGV.shift
    break if arg == '--'
    TkORCA_SERVICE_OPTIONS << arg
  end
end

TkORCA_SERVICE_APPLICATIONS = ARGV.dup
=end


##############################################
#  show version ?  -->  print and exit
##############################################
if OPTION['version']
  load(File.join(SERVER_DIR, 'show_version.rb'))
  Kernel.exit(0)
end


########################################################
#  setup SERVER_PARAM
########################################################
SERVER_PARAM = {}

def setup_SERVER_PARAMS
  SERVER_PARAM.clear

  #  hostname
  SERVER_PARAM['hostname'] = Socket.gethostname
  SERVER_PARAM['shortname'] = SERVER_PARAM['hostname'].split('.')[0]

  # hangup signal
  if OPTION['hangup_signal']
    SERVER_PARAM['hangup_signal'] = OPTION['hangup_signal']
  else
    SERVER_PARAM['hangup_signal'] = TkORCA_SIGHUP
  end
  unless SERVER_PARAM['hangup_signal']
    raise RuntimeError, "Fail to select proper singal for TkORCA_SIGHUP." + 
      " It is required for interactive control of application processes." + 
      " Please define TkORCA_SIGHUP on `server_base_defs' file" + 
      " or use `-S' (or `--hangup_signal') option."
  end

  #  null device
  if OPTION['null_device']
    SERVER_PARAM['null_device'] = OPTION['null_device']
  else
    SERVER_PARAM['null_device'] = NULL_DEVICE
  end

  #  ruby binary
  if OPTION['ruby_bin']
    SERVER_PARAM['ruby_bin'] = File.expand_path(OPTION['ruby_bin'])
  end

  #  user/group
  if OPTION['user']
    SERVER_PARAM['user'] = OPTION['user']
  else
    SERVER_PARAM['user'] = TkORCA_USER
  end

  if OPTION['group']
    SERVER_PARAM['group'] = OPTION['group']
  else
    SERVER_PARAM['group'] = TkORCA_GROUP
  end

  if OPTION['server_user']
    SERVER_PARAM['server_user'] = OPTION['server_user']
  else
    SERVER_PARAM['server_user'] = TkORCA_SERVER_USER
  end

  if OPTION['server_group']
    SERVER_PARAM['server_group'] = OPTION['server_group']
  else
    SERVER_PARAM['server_group'] = TkORCA_SERVER_GROUP
  end

  #  TCP param (hostname/port/listen/controller_port/reverse_lookup)
  if OPTION['hostaddr']
    SERVER_PARAM['hostaddr'] = OPTION['hostaddr']
  else
    SERVER_PARAM['hostaddr'] = HOST_ADDRESS
  end

  if OPTION['port'] >= 0
    SERVER_PARAM['port'] = OPTION['port']
  else
    SERVER_PARAM['port'] = SERVICE_PORT
  end

  if OPTION['listen'] > 0
    SERVER_PARAM['listen'] = OPTION['listen']
  else
    SERVER_PARAM['listen'] = LISTEN_BACKLOG
  end

  if OPTION['control_port'] >= 0
    SERVER_PARAM['control_port'] = OPTION['control_port']
  elsif Object.const_defined?(:CONTROL_PORT)
    SERVER_PARAM['control_port'] = CONTROL_PORT
  else
    SERVER_PARAM['control_port'] = SERVER_PARAM['port'] + 1
  end

  if OPTION['interactive']
    SERVER_PARAM['interactive'] = (OPTION['interactive'])? true: false
  else
    SERVER_PARAM['interactive'] = TkORCA_INTERACTIVE
  end

  if OPTION['reverse_lookup']
    SERVER_PARAM['reverse_lookup'] = OPTION['reverse_lookup']
  else
    SERVER_PARAM['reverse_lookup'] = REVERSE_LOOKUP
  end

  if OPTION['use_tmp_home']
    SERVER_PARAM['use_tmp_home'] = (OPTION['use_tmp_home'])? true: false
  else
    SERVER_PARAM['use_tmp_home'] = USE_TEMP_HOME_DIRECTORY
  end

  #  server_rcfile
  if OPTION['serverrc']
    SERVER_PARAM['server_rcfile'] = File.expand_path(OPTION['serverrc'])
  elsif SERVER_RCFILE
    SERVER_PARAM['server_rcfile'] = File.expand_path(SERVER_RCFILE)
  else
    SERVER_PARAM['server_rcfile'] = nil
  end

  #  others
  if OPTION['tmp_dir']
    SERVER_PARAM['tmp_dir'] = File.expand_path(OPTION['tmp_dir'])
  elsif Object.const_defined?(:SERVER_TMPDIR)
    SERVER_PARAM['tmp_dir'] = File.expand_path(SERVER_TMPDIR)
  else
    SERVER_PARAM['tmp_dir'] = File.expand_path(Dir.tmpdir)
  end

  if OPTION['tmp_home_dir']
    SERVER_PARAM['tmp_home_parent'] = File.expand_path(OPTION['tmp_home_dir'])
  elsif Object.const_defined?(:SERVER_TMP_HOME_DIR)
    SERVER_PARAM['tmp_home_parent'] = File.expand_path(SERVER_TMP_HOME_DIR)
  end

  if OPTION['app_root_dir']
    SERVER_PARAM['app_root_dir'] = File.expand_path(OPTION['app_root_dir'])
  elsif Object.const_defined?(:APPLICATION_ROOT_DIR)
    SERVER_PARAM['app_root_dir'] = File.expand_path(APPLICATION_ROOT_DIR)
  end

  if OPTION['tkorca_path']
    SERVER_PARAM['tkorca_path'] = File.expand_path(OPTION['tkorca_path'])
  else
    SERVER_PARAM['tkorca_path'] = File.join(SERVER_DIR, 'tkorca.rb')
  end

  if OPTION['xauth_path']
    SERVER_PARAM['xauth_path'] = File.expand_path(OPTION['xauth_path'])
  end

  if OPTION['xauth_dir']
    SERVER_PARAM['xauth_dir'] = File.expand_path(OPTION['xauth_dir'])
  end

  if OPTION['xvnc_path']
    SERVER_PARAM['xvnc_path'] = File.expand_path(OPTION['xvnc_path'])
  end

  if OPTION['control_passwd']
    SERVER_PARAM['control_passwd'] = File.expand_path(OPTION['control_passwd'])
  else
    SERVER_PARAM['control_passwd'] = File.expand_path(CONTROL_PASSWD_FILE)
  end

  if OPTION['rfbauth_passwd']
    SERVER_PARAM['rfbauth_passwd'] = File.expand_path(OPTION['rfbauth_passwd'])
  elsif RFBAUTH_PASSWD_FILE
    SERVER_PARAM['rfbauth_passwd'] = File.expand_path(RFBAUTH_PASSWD_FILE)
  else
    SERVER_PARAM['rfbauth_passwd'] = nil
  end

  if OPTION['desktop_name']
    SERVER_PARAM['desktop_name'] = OPTION['desktop_name']
  else
    SERVER_PARAM['desktop_name'] = nil
  end

  if OPTION['color_depth']
    SERVER_PARAM['color_depth'] = OPTION['color_depth']
  else
    SERVER_PARAM['color_depth'] = nil
  end

  if OPTION['wall_size']
    SERVER_PARAM['wall_size'] = OPTION['wall_size']
  else
    SERVER_PARAM['wall_size'] = nil
  end

  if OPTION['deferupdate']
    SERVER_PARAM['deferupdate'] = OPTION['deferupdate']
  else
    SERVER_PARAM['deferupdate'] = nil
  end
end

setup_SERVER_PARAMS()


########################################################
#  write PID-file
########################################################
#if File.exist?(SERVER_PID)
#  raise RuntimeError, "PID file `#{SERVER_PID}' exists." + 
#    ' Another Ruby/TkORCA server process may be running.'
#end
if Object.const_defined?(:SERVER_PID) && SERVER_PID
  SERVER_PARAM['server_pid'] = File.expand_path(SERVER_PID)
else
  SERVER_PARAM['server_pid'] = File.join(SERVER_PARAM['tmp_dir'], 'tkorca.pid')
end
pid_file = File.open(SERVER_PARAM['server_pid'], 'a')
unless pid_file.flock(File::LOCK_EX|File::LOCK_NB)
  raise RuntimeError, "fail to lock PID file `#{SERVER_PARAM['server_pid']}'."+     ' Another Ruby/TkORCA server process may be running.'
  exit!(1)
end
pid_file.truncate(0)
pid_file.puts Process.pid
pid_file.flush


########################################################
# decide ruby binary path
########################################################
unless SERVER_PARAM['ruby_bin']
  if Object.const_defined?(:RUBY_BINARY)
    SERVER_PARAM['ruby_bin'] = File.expand_path(RUBY_BINARY)
  else
    require 'rbconfig'
    bin_name = Config::CONFIG['rubyw_install_name']
    if !bin_name || bin_name.empty?
      bin_name = Config::CONFIG['ruby_install_name']
    end

    if FileTest.exist?(path = File.join(Config::CONFIG['bindir'], bin_name))
      SERVER_PARAM['ruby_bin'] = path
    elsif FileTest.exist?(path = File.join('/usr/local/bin', bin_name))
      SERVER_PARAM['ruby_bin'] = path
    elsif FileTest.exist?(path = File.join('/usr/bin', bin_name))
      SERVER_PARAM['ruby_bin'] = path
    else # entrust command-search-path
      SERVER_PARAM['ruby_bin'] = 
        `which #{bin_name} 2>#{SERVER_PARAM['null_device']}`.chomp
      if $?.exitstatus != 0
        # Pray to the gods ...
        SERVER_PARAM['ruby_bin'] = bin_name
      end
    end
  end
end


########################################################
#  require components
########################################################
require 'server/logger.rb'
require 'server/tkorca_accepter.rb'
require 'server/tkorca_controller.rb'


########################################################
#  setup logfile
########################################################
logfile = nil
if OPTION['logfile']
  logfile = File.expand_path(OPTION['logfile'])
elsif Object.const_defined?(:SERVER_LOGFILE) && SERVER_LOGFILE
  logfile = File.expand_path(SERVER_LOGFILE)
else
  logfile = File.join(SERVER_PARAM['tmp_dir'], 'tkorca.log')
end
SERVER_PARAM['logfile'] = logfile

LOGGER = TkORCA_LOGGER.new(SERVER_PARAM['logfile'])

def LOGGER.startlog(init=false)
  self.log("#{(init)? '': 're'}start Ruby/TkORCA", 
           "version #{TkORCA_VERSION} (#{TkORCA_RELEASE_DATE})", 
           "on #{SERVER_PARAM['ruby_bin']}", 
           "#{RUBY_VERSION} (#{RUBY_RELEASE_DATE}) [#{RUBY_PLATFORM}]")
  self.log("server user/group =", 
           "#{SERVER_PARAM['server_user']}/#{SERVER_PARAM['server_group']}")
  self.log("client user/group =", 
           "#{SERVER_PARAM['user']}/#{SERVER_PARAM['group']}")
end


########################################################
#  should be do when server params are updated 
########################################################
def server_param_update_operation
  BasicSocket.do_not_reverse_lookup = ! SERVER_PARAM['reverse_lookup']
end

server_param_update_operation()

########################################################
#  keep server param
########################################################
ORIG_SERVER_PARAM = SERVER_PARAM.dup


########################################################
#  setup NULL I/O
########################################################
TkORCA_NULL_IO = open(SERVER_PARAM['null_device'], 'a+')


########################################################
#  write startup message to logfile
########################################################
LOGGER.startlog(true)


########################################################
#  STDERR send to STDOUT
########################################################
$stderr.reopen($stdout)


########################################################
#  load server_rcfile (if exist)
########################################################
def load_server_rcfile
  rcfile = SERVER_PARAM['server_rcfile']
  if rcfile && FileTest.readable?(rcfile)
    LOGGER.log("load server default rcfile `#{rcfile}'")
    begin
      load(rcfile)
    rescue Exception => e
      LOGGER.log("ERROR: #{e.class}: #{e.message}")
      LOGGER.log("fail to load server default rcfile `#{rcfile}'")
      raise e
    end
  else
    LOGGER.log("cannot read rcfile `#{rcfile}' -- ignore")
  end
end

load_server_rcfile()


########################################################
#  change NULL I/O ?
########################################################
if SERVER_PARAM['null_device'] != ORIG_SERVER_PARAM['null_device']
  # changed in serverrc
  TkORCA_NULL_IO.reopen(SERVER_PARAM['null_device'], 'a+')
end


########################################################
#  setup trap operation
########################################################
exit_proc = proc{
  pid_file.flock(File::LOCK_UN) rescue nil
  pid_file.close rescue nil
  # File.delete(SERVER_PID) rescue nil
  File.delete(SERVER_PARAM['server_pid']) rescue nil
  TkORCA_Accepter.shutdown rescue nil
  TkORCA_Controller.kill rescue nil
  TkORCA_Service.kill_all_connections rescue nil
  LOGGER.log("exit Ruby/TkORCA server") rescue nil
  LOGGER.close rescue nil
}
# STDERR.puts exit_proc.inspect; STDERR.flush
at_exit(&exit_proc)

# restart trap
Signal.trap(SERVER_PARAM['hangup_signal']) {
  LOGGER.restart rescue nil
  LOGGER.log("restart Ruby/TkORCA server") rescue nil
  LOGGER.log(" ... reset server parameters") rescue nil

  null_dev = SERVER_PARAM['null_device']

  SERVER_PARAM.clear rescue nil
  SERVER_PARAM.update(ORIG_SERVER_PARAM) rescue nil
  LOGGER.log(" ... re-load (new) serverrc file") rescue nil
  load_server_rcfile

  if SERVER_PARAM['null_device'] != null_dev
    TkORCA_NULL_IO.reopen(SERVER_PARAM['null_device'], 'a+')
  end

  server_param_update_operation()

  LOGGER.log(" ... restart accepter") rescue nil
  TkORCA_Accepter.restart

  LOGGER.log(" ... restart controller") rescue nil
  TkORCA_Controller.restart

  LOGGER.log("restart operation is complete") rescue nil
}

# exit trap :: call at_exit operation by myself
Signal.trap(:INT)  { exit_proc.call; Kernel.exit!(0) }
Signal.trap(:TERM) { exit_proc.call; Kernel.exit!(0) }


########################################################
#  start accepter
########################################################
TkORCA_Accepter.start


########################################################
#  start controller thread
########################################################
TkORCA_Controller.start


########################################################
#  wait forever
########################################################
begin
  sleep
rescue Exception
end
